import greenfoot.*;
/**
 * Write a description of class AI here.
 * 
 * @author Taylor Born
 * @version November 2010
 */
import java.util.*;

class TableEntry
{
    public int score;
    public int depth = -1;
    public long lock;
    public int age;
}
class Move
{
    public Move next = null;
    public long from, to, jumpPawn;// occupied, bOccupied, wOccupied,
    public long[] black, white;
    public int type;
    public int promotion = -1;
    public boolean castled;
    
    public Move(long from, long to, long[] b, long[] w, long jp, int t) // long occ, long bOcc, long wOcc,
    {
        this.from = from;
        this.to = to;
//         occupied = occ;
//         bOccupied = bOcc;
//         wOccupied = wOcc;
        black = b;
        white = w;
        jumpPawn = jp;
        type = t;
    }
}
class MoveList
{
    public Move first;

    public void add(Move m)
    {
        if (first == null)
            first = m;
        else if (m.type <= first.type)
        {
            Move temp = first;
            first = m;
            first.next = temp;
        }
        else if (first.next == null)
            first.next = m;
        else
        {
            Move mm = first;
            while (mm.next != null)
            {
                if (mm.next.type >= m.type)
                {
                    Move temp = mm.next;
                    mm.next = m;
                    m.next = temp;
                    return;
                }
                mm = mm.next;
            }
            mm.next = m;
        }
    }
}
public class AI  extends Actor
{
    static int clock = 0;
    static TableEntry[] table = new TableEntry[131072];

    // Translate data structures and calculate
    public AI(ArrayList<ArrayList<Square>> grid, boolean bCastled, boolean bRA, boolean bRB, boolean bK, boolean wCastled, boolean wRA, boolean wRB, boolean wK, Point jP)
    {
        finished = false;
        long blackOccupied = 0;
        long[] blackPieces = new long[6];
        long whiteOccupied = 0;
        long[] whitePieces = new long[6];
        for (int r = 0; r < 8; r++)
            for (int c = 0; c < 8; c++)
            {
                Piece p = grid.get(r).get(c).getPiece();
                if (p != null)
                {
                    if (p.getPlayer() == 1)
                    {
                        whiteOccupied |= (1L << ((7 - r) * 8 + 7 - c));
                        whitePieces[p.getType()] |= (1L << ((7 - r) * 8 + 7 - c));
                    }
                    else
                    {
                        blackOccupied |= (1L << ((7 - r) * 8 + 7 - c));
                        blackPieces[p.getType()] |= (1L << ((7 - r) * 8 + 7 - c));
                    }
                }
            }
        long jumpPawn = 0;
        if (jP != null)
            jumpPawn |= (1L << ((7 - jP.getY()) * 8 + 7 - jP.getX()));
        calculate(0, blackOccupied | whiteOccupied, blackOccupied, whiteOccupied, blackPieces, whitePieces, bCastled, bRA, bRB, bK, wCastled, wRA, wRB, wK, jumpPawn, null);
        finished = true;
        Greenfoot.setSpeed(50);
    }
    boolean finished;
    private long bestFrm;
    private long best2;
    private int promoPawn;
    public Point getBestFrom()
    {
        clock++;
        int n = Long.numberOfLeadingZeros(bestFrm);
        return new Point(n % 8, n / 8);
    }
    public Point getBestTo()
    {
        int n = Long.numberOfLeadingZeros(best2);
        return new Point(n % 8, n / 8);
    }
    public int getPromotion()
    {
        return promoPawn;
    }
    private Integer calculate(int depth, long occupied, long blackOccupied, long whiteOccupied, long[] blackPieces, long[] whitePieces, boolean bCastled, boolean bRA, boolean bRB, boolean bK, boolean wCastled, boolean wRA, boolean wRB, boolean wK, long jumpPawn, Integer alpha)
    {
        Integer score = null;
        if ((blackPieces[5] & squareBits[59]) == 0)
            bK = false;
        if ((blackPieces[1] & squareBits[63]) == 0)
            bRA = false;
        if ((blackPieces[1] & squareBits[56]) == 0)
            bRB = false;
        if ((whitePieces[5] & squareBits[3]) == 0)
            wK = false;
        if ((whitePieces[1] & squareBits[7]) == 0)
            wRA = false;
        if ((whitePieces[1] & squareBits[0]) == 0)
            wRB = false;
        if (depth == 5)
        {
            score = 0;
            score += Long.bitCount(blackPieces[0]) * 101;
            score += Long.bitCount(blackPieces[1]) * 501;
            score += Long.bitCount(blackPieces[2]) * 301;
            score += Long.bitCount(blackPieces[3]) * 326;
            score += Long.bitCount(blackPieces[4]) * 901;
            score += Long.bitCount(blackPieces[5]) * 9001;
            score -= Long.bitCount(whitePieces[0]) * 100;
            score -= Long.bitCount(whitePieces[1]) * 500;
            score -= Long.bitCount(whitePieces[2]) * 300;
            score -= Long.bitCount(whitePieces[3]) * 325;
            score -= Long.bitCount(whitePieces[4]) * 900;
            score -= Long.bitCount(whitePieces[5]) * 9000;
            for (int i = 1; i < 5; i++)
            {
                score -= Long.bitCount(areaAroundKing[Long.numberOfTrailingZeros(blackPieces[5])] & whitePieces[i]) * i;
                score += Long.bitCount(areaAroundKing[Long.numberOfTrailingZeros(whitePieces[5])] & blackPieces[i]) * i;
                score -= Long.bitCount(middleOfBoard & whitePieces[i]) * i;
                score += Long.bitCount(middleOfBoard & blackPieces[i]) * i;
            }
            if (bCastled)
                score += 10;
            else if (bK)
            {
                if (bRA && bRB)
                    score -= 24;
                else if (bRA)
                    score -= 40;
                else if (bRB)
                    score -= 80;
                else
                    score -= 120;
            }
            else
                score -= 120;
            if (wCastled)
                score -= 10;
            else if (wK)
            {
                if (wRA && wRB)
                    score += 24;
                else if (wRA)
                    score += 40;
                else if (wRB)
                    score += 80;
                else
                    score += 120;
            }
            else
                score += 120;
            return score;
        }
        long bestFrom = -1;
        long bestTo = -1;
        long l;
        Integer s;
        Integer his;
        MoveList myMoves = new MoveList();
        if (depth % 2 == 0) // black
        {
            for (int i = 0; i < 6; i++)
            {
                switch (i)
                {
                    case 0:     l = blackPieces[0];
                                while (l != 0)
                                {
                                    long t = Long.highestOneBit(l);
                                    int trailing = Long.numberOfTrailingZeros(t);
                                    long m = blackPawnAttacks[trailing] & whiteOccupied;
                                    while (m != 0)
                                    {
                                        long tt = Long.highestOneBit(m);
                                        long[] wPieces = new long[6];
                                        for (int p = 0; p < 6; p++)
                                            wPieces[p] = (whitePieces[p] & tt) != 0 ? whitePieces[p] ^ tt : whitePieces[p];
                                        long[] bPieces = new long[6];
                                        bPieces[0] = blackPieces[0] ^ t;
                                        for (int p = 1; p < 6; p++)
                                            bPieces[p] = blackPieces[p];
                                        if (Long.numberOfTrailingZeros(tt) < 9)
                                        {
                                            long[] bbPieces = new long[6];
                                            bbPieces[0] = bPieces[0];
                                            for (int p = 1; p < 5; p++)
                                            {
                                                for (int pp = 1; pp < 6; pp++)
                                                    if (pp == p)
                                                        bbPieces[pp] = blackPieces[pp] | tt;
                                                    else
                                                        bbPieces[pp] = blackPieces[pp];
//                                                 bOcc = blackOccupied ^ t | tt;
//                                                 wOcc = whiteOccupied ^ tt;
//                                                 occ = bOcc | wOcc;
                                                Move move = new Move(t, tt, bbPieces, wPieces, 0, -6); // -6 because a capture and a promotion
                                                move.promotion = p;
                                                myMoves.add(move);
                                            }
                                        }
                                        else
                                        {
                                            bPieces[0] |= tt;
//                                             bOcc = blackOccupied ^ t | tt;
//                                             wOcc = whiteOccupied ^ tt;
//                                             occ = bOcc | wOcc;
                                            int pieceC = 0;
                                            for (pieceC = 0; pieceC < 5; pieceC++)
                                                if ((wPieces[pieceC] & tt) != 0)
                                                    break;
                                            myMoves.add(new Move(t, tt, bPieces, wPieces, 0, -pieceC));
                                        }
                                        m ^= tt;
                                    }
                                    m = blackPawnMoves[trailing];
                                    long tt = Long.highestOneBit(m);
                                    if ((tt & ~occupied) != 0)
                                    {
                                        long[] bPieces = new long[6];
                                        bPieces[0] = blackPieces[0] ^ t;
                                        for (int p = 1; p < 6; p++)
                                            bPieces[p] = blackPieces[p];
                                        if (Long.numberOfTrailingZeros(tt) < 9)
                                        {
                                            long[] bbPieces = new long[6];
                                            bbPieces[0] = bPieces[0];
                                            for (int p = 1; p < 5; p++)
                                            {
                                                for (int pp = 1; pp < 6; pp++)
                                                    if (pp == p)
                                                        bbPieces[pp] = blackPieces[pp] | tt;
                                                    else
                                                        bbPieces[pp] = blackPieces[pp];
//                                                 bOcc = blackOccupied ^ t | tt;
//                                                 occ = bOcc | whiteOccupied;
                                                Move move = new Move(t, tt, bbPieces, whitePieces, 0, 1);
                                                move.promotion = p;
                                                myMoves.add(move);
                                            }
                                        }
                                        else
                                        {
                                            bPieces[0] |= tt;
//                                             bOcc = blackOccupied ^ t | tt;
//                                             occ = bOcc | whiteOccupied;
                                            myMoves.add(new Move(t, tt, bPieces, whitePieces, 0, 5));
                                            m ^= tt;
                                            long secondMove = Long.highestOneBit(m);
                                            if ((secondMove & ~occupied) != 0)
                                            {
                                                bPieces[0] = blackPieces[0] ^ t | secondMove;
//                                                 bOcc = blackOccupied ^ t | secondMove;
//                                                 occ = bOcc | whiteOccupied;
                                                myMoves.add(new Move(t, secondMove, bPieces, whitePieces, tt, 5));
                                            }
                                        }
                                    }
                                    tt = blackPawnAttacks[trailing] & jumpPawn;
                                    if (tt != 0)
                                    {
                                        long pn = squareBits[Long.numberOfTrailingZeros(jumpPawn) - 8];
                                        long[] wPieces = new long[6];
                                        wPieces[0] = whitePieces[0] ^ pn;
                                        for (int p = 1; p < 6; p++)
                                            wPieces[p] = whitePieces[p];
                                        long[] bPieces = new long[6];
                                        for (int p = 1; p < 6; p++)
                                            bPieces[p] = blackPieces[p];
                                        bPieces[0] = blackPieces[0] ^ t | tt;
//                                         bOcc = blackOccupied ^ t | tt;
//                                         wOcc = whiteOccupied ^ pn;
//                                         occ = bOcc | wOcc;
                                        myMoves.add(new Move(t, tt, bPieces, wPieces, 0, 0));
                                    }
                                    l ^= t;
                                }
                                break;
                    case 4:
                    case 1:     l = blackPieces[i];
                                while (l != 0)
                                {
                                    long t = Long.highestOneBit(l);
                                    int trailing = Long.numberOfTrailingZeros(t);
                                    for (long lng : slideSouth.get(trailing))
                                        if ((occupied & lng) != 0)
                                        {
                                            if ((whiteOccupied & lng) != 0)
                                            {
                                                long[] wPieces = new long[6];
                                                int pieceC = 0;
                                                for (int p = 0; p < 6; p++)
                                                    if ((whitePieces[p] & lng) != 0)
                                                    {
                                                        pieceC = p;
                                                        wPieces[p] = whitePieces[p] ^ lng;
                                                    }
                                                    else
                                                        wPieces[p] = whitePieces[p];
                                                long[] bPieces = new long[6];
                                                for (int p = 0; p < 6; p++)
                                                    if (p == i)
                                                        bPieces[p] = blackPieces[p] ^ t | lng;
                                                    else
                                                        bPieces[p] = blackPieces[p];
//                                                 bOcc = blackOccupied ^ t | lng;
//                                                 wOcc = whiteOccupied ^ lng;
//                                                 occ = bOcc | wOcc;
                                                myMoves.add(new Move(t, lng, bPieces, wPieces, 0, -pieceC));
                                            }
                                            break;
                                        }
                                        else
                                        {
                                            long[] bPieces = new long[6];
                                            for (int p = 0; p < 6; p++)
                                                if (p == i)
                                                    bPieces[p] = blackPieces[p] ^ t | lng;
                                                else
                                                    bPieces[p] = blackPieces[p];
//                                             bOcc = blackOccupied ^ t | lng;
//                                             occ = bOcc | whiteOccupied;
                                            myMoves.add(new Move(t, lng, bPieces, whitePieces, 0, 5));
                                        }
                                    for (long lng : slideEast.get(trailing))
                                        if ((occupied & lng) != 0)
                                        {
                                            if ((whiteOccupied & lng) != 0)
                                            {
                                                long[] wPieces = new long[6];
                                                int pieceC = 0;
                                                for (int p = 0; p < 6; p++)
                                                    if ((whitePieces[p] & lng) != 0)
                                                    {
                                                        pieceC = p;
                                                        wPieces[p] = whitePieces[p] ^ lng;
                                                    }
                                                    else
                                                        wPieces[p] = whitePieces[p];
                                                long[] bPieces = new long[6];
                                                for (int p = 0; p < 6; p++)
                                                    if (p == i)
                                                        bPieces[p] = blackPieces[p] ^ t | lng;
                                                    else
                                                        bPieces[p] = blackPieces[p];
//                                                 bOcc = blackOccupied ^ t | lng;
//                                                 wOcc = whiteOccupied ^ lng;
//                                                 occ = bOcc | wOcc;
                                                myMoves.add(new Move(t, lng, bPieces, wPieces, 0, -pieceC));
                                            }
                                            break;
                                        }
                                        else
                                        {
                                            long[] bPieces = new long[6];
                                            for (int p = 0; p < 6; p++)
                                                if (p == i)
                                                    bPieces[p] = blackPieces[p] ^ t | lng;
                                                else
                                                    bPieces[p] = blackPieces[p];
//                                             bOcc = blackOccupied ^ t | lng;
//                                             occ = bOcc | whiteOccupied;
                                            myMoves.add(new Move(t, lng, bPieces, whitePieces, 0, 5));
                                        }
                                    for (long lng : slideWest.get(trailing))
                                        if ((occupied & lng) != 0)
                                        {
                                            if ((whiteOccupied & lng) != 0)
                                            {
                                                long[] wPieces = new long[6];
                                                int pieceC = 0;
                                                for (int p = 0; p < 6; p++)
                                                    if ((whitePieces[p] & lng) != 0)
                                                    {
                                                        pieceC = p;
                                                        wPieces[p] = whitePieces[p] ^ lng;
                                                    }
                                                    else
                                                        wPieces[p] = whitePieces[p];
                                                long[] bPieces = new long[6];
                                                for (int p = 0; p < 6; p++)
                                                    if (p == i)
                                                        bPieces[p] = blackPieces[p] ^ t | lng;
                                                    else
                                                        bPieces[p] = blackPieces[p];
//                                                 bOcc = blackOccupied ^ t | lng;
//                                                 wOcc = whiteOccupied ^ lng;
//                                                 occ = bOcc | wOcc;
                                                myMoves.add(new Move(t, lng, bPieces, wPieces, 0, -pieceC));
                                            }
                                            break;
                                        }
                                        else
                                        {
                                            long[] bPieces = new long[6];
                                            for (int p = 0; p < 6; p++)
                                                if (p == i)
                                                    bPieces[p] = blackPieces[p] ^ t | lng;
                                                else
                                                    bPieces[p] = blackPieces[p];
//                                             bOcc = blackOccupied ^ t | lng;
//                                             occ = bOcc | whiteOccupied;
                                            myMoves.add(new Move(t, lng, bPieces, whitePieces, 0, 5));
                                        }
                                    for (long lng : slideNorth.get(trailing))
                                        if ((occupied & lng) != 0)
                                        {
                                            if ((whiteOccupied & lng) != 0)
                                            {
                                                long[] wPieces = new long[6];
                                                int pieceC = 0;
                                                for (int p = 0; p < 6; p++)
                                                    if ((whitePieces[p] & lng) != 0)
                                                    {
                                                        pieceC = p;
                                                        wPieces[p] = whitePieces[p] ^ lng;
                                                    }
                                                    else
                                                        wPieces[p] = whitePieces[p];
                                                long[] bPieces = new long[6];
                                                for (int p = 0; p < 6; p++)
                                                    if (p == i)
                                                        bPieces[p] = blackPieces[p] ^ t | lng;
                                                    else
                                                        bPieces[p] = blackPieces[p];
//                                                 bOcc = blackOccupied ^ t | lng;
//                                                 wOcc = whiteOccupied ^ lng;
//                                                 occ = bOcc | wOcc;
                                                myMoves.add(new Move(t, lng, bPieces, wPieces, 0, -pieceC));
                                            }
                                            break;
                                        }
                                        else
                                        {
                                            long[] bPieces = new long[6];
                                            for (int p = 0; p < 6; p++)
                                                if (p == i)
                                                    bPieces[p] = blackPieces[p] ^ t | lng;
                                                else
                                                    bPieces[p] = blackPieces[p];
//                                             bOcc = blackOccupied ^ t | lng;
//                                             occ = bOcc | whiteOccupied;
                                            myMoves.add(new Move(t, lng, bPieces, whitePieces, 0, 5));
                                        }
                                    l ^= t;
                                }
                                if (i == 1)
                                    break;
                    case 3:     l = blackPieces[i];
                                while (l != 0)
                                {
                                    long t = Long.highestOneBit(l);
                                    int trailing = Long.numberOfTrailingZeros(t);
                                    for (long lng : slideNW.get(trailing))
                                        if ((occupied & lng) != 0)
                                        {
                                            if ((whiteOccupied & lng) != 0)
                                            {
                                                long[] wPieces = new long[6];
                                                int pieceC = 0;
                                                for (int p = 0; p < 6; p++)
                                                {
                                                    if ((whitePieces[p] & lng) != 0)
                                                    {
                                                        pieceC = p;
                                                        wPieces[p] = whitePieces[p] ^ lng;
                                                    }
                                                    else
                                                        wPieces[p] = whitePieces[p];
                                                }
                                                long[] bPieces = new long[6];
                                                for (int p = 0; p < 6; p++)
                                                    if (p == i)
                                                        bPieces[p] = blackPieces[p] ^ t | lng;
                                                    else
                                                        bPieces[p] = blackPieces[p];
//                                                 bOcc = blackOccupied ^ t | lng;
//                                                 wOcc = whiteOccupied ^ lng;
//                                                 occ = bOcc | wOcc;
                                                myMoves.add(new Move(t, lng, bPieces, wPieces, 0, -pieceC));
                                            }
                                            break;
                                        }
                                        else
                                        {
                                            long[] bPieces = new long[6];
                                            for (int p = 0; p < 6; p++)
                                                if (p == i)
                                                    bPieces[p] = blackPieces[p] ^ t | lng;
                                                else
                                                    bPieces[p] = blackPieces[p];
//                                             bOcc = blackOccupied ^ t | lng;
//                                             occ = bOcc | whiteOccupied;
                                            myMoves.add(new Move(t, lng, bPieces, whitePieces, 0, 5));
                                        }
                                    for (long lng : slideNE.get(trailing))
                                        if ((occupied & lng) != 0)
                                        {
                                            if ((whiteOccupied & lng) != 0)
                                            {
                                                long[] wPieces = new long[6];
                                                int pieceC = 0;
                                                for (int p = 0; p < 6; p++)
                                                {
                                                    if ((whitePieces[p] & lng) != 0)
                                                    {
                                                        pieceC = p;
                                                        wPieces[p] = whitePieces[p] ^ lng;
                                                    }
                                                    else
                                                        wPieces[p] = whitePieces[p];
                                                }
                                                long[] bPieces = new long[6];
                                                for (int p = 0; p < 6; p++)
                                                    if (p == i)
                                                        bPieces[p] = blackPieces[p] ^ t | lng;
                                                    else
                                                        bPieces[p] = blackPieces[p];
//                                                 bOcc = blackOccupied ^ t | lng;
//                                                 wOcc = whiteOccupied ^ lng;
//                                                 occ = bOcc | wOcc;
                                                myMoves.add(new Move(t, lng, bPieces, wPieces, 0, -pieceC));
                                            }
                                            break;
                                        }
                                        else
                                        {
                                            long[] bPieces = new long[6];
                                            for (int p = 0; p < 6; p++)
                                                if (p == i)
                                                    bPieces[p] = blackPieces[p] ^ t | lng;
                                                else
                                                    bPieces[p] = blackPieces[p];
//                                             bOcc = blackOccupied ^ t | lng;
//                                             occ = bOcc | whiteOccupied;
                                            myMoves.add(new Move(t, lng, bPieces, whitePieces, 0, 5));
                                        }
                                    for (long lng : slideSW.get(trailing))
                                        if ((occupied & lng) != 0)
                                        {
                                            if ((whiteOccupied & lng) != 0)
                                            {
                                                long[] wPieces = new long[6];
                                                int pieceC = 0;
                                                for (int p = 0; p < 6; p++)
                                                {
                                                    if ((whitePieces[p] & lng) != 0)
                                                    {
                                                        pieceC = p;
                                                        wPieces[p] = whitePieces[p] ^ lng;
                                                    }
                                                    else
                                                        wPieces[p] = whitePieces[p];
                                                }
                                                long[] bPieces = new long[6];
                                                for (int p = 0; p < 6; p++)
                                                    if (p == i)
                                                        bPieces[p] = blackPieces[p] ^ t | lng;
                                                    else
                                                        bPieces[p] = blackPieces[p];
//                                                 bOcc = blackOccupied ^ t | lng;
//                                                 wOcc = whiteOccupied ^ lng;
//                                                 occ = bOcc | wOcc;
                                                myMoves.add(new Move(t, lng, bPieces, wPieces, 0, -pieceC));
                                            }
                                            break;
                                        }
                                        else
                                        {
                                            long[] bPieces = new long[6];
                                            for (int p = 0; p < 6; p++)
                                                if (p == i)
                                                    bPieces[p] = blackPieces[p] ^ t | lng;
                                                else
                                                    bPieces[p] = blackPieces[p];
//                                             bOcc = blackOccupied ^ t | lng;
//                                             occ = bOcc | whiteOccupied;
                                            myMoves.add(new Move(t, lng, bPieces, whitePieces, 0, 5));
                                        }
                                    for (long lng : slideSE.get(trailing))
                                        if ((occupied & lng) != 0)
                                        {
                                            if ((whiteOccupied & lng) != 0)
                                            {
                                                long[] wPieces = new long[6];
                                                int pieceC = 0;
                                                for (int p = 0; p < 6; p++)
                                                {
                                                    if ((whitePieces[p] & lng) != 0)
                                                    {
                                                        pieceC = p;
                                                        wPieces[p] = whitePieces[p] ^ lng;
                                                    }
                                                    else
                                                        wPieces[p] = whitePieces[p];
                                                }
                                                long[] bPieces = new long[6];
                                                for (int p = 0; p < 6; p++)
                                                    if (p == i)
                                                        bPieces[p] = blackPieces[p] ^ t | lng;
                                                    else
                                                        bPieces[p] = blackPieces[p];
//                                                 bOcc = blackOccupied ^ t | lng;
//                                                 wOcc = whiteOccupied ^ lng;
//                                                 occ = bOcc | wOcc;
                                                myMoves.add(new Move(t, lng, bPieces, wPieces, 0, -pieceC));
                                            }
                                            break;
                                        }
                                        else
                                        {
                                            long[] bPieces = new long[6];
                                            for (int p = 0; p < 6; p++)
                                                if (p == i)
                                                    bPieces[p] = blackPieces[p] ^ t | lng;
                                                else
                                                    bPieces[p] = blackPieces[p];
//                                             bOcc = blackOccupied ^ t | lng;
//                                             occ = bOcc | whiteOccupied;
                                            myMoves.add(new Move(t, lng, bPieces, whitePieces, 0, 5));
                                        }
                                    l ^= t;
                                }
                                break;
                    case 2:     l = blackPieces[2];
                                while (l != 0)
                                {
                                    long t = Long.highestOneBit(l);
                                    int trailing = Long.numberOfTrailingZeros(t);
                                    long m = knightMoves[trailing] & ~blackOccupied;
                                    while (m != 0)
                                    {
                                        long tt = Long.highestOneBit(m);
                                        long[] wPieces = new long[6];
                                        int pieceC = 5;
                                        for (int p = 0; p < 6; p++)
                                        {
                                            if ((whitePieces[p] & tt) != 0)
                                            {
                                                pieceC = p;
                                                wPieces[p] = whitePieces[p] ^ tt;
                                            }
                                            else
                                                wPieces[p] = whitePieces[p];
                                        }
                                        long[] bPieces = new long[6];
                                        for (int p = 0; p < 6; p++)
                                            bPieces[p] = blackPieces[p];
                                        bPieces[2] = bPieces[2] ^ t | tt;
//                                         bOcc = blackOccupied ^ t | tt;
//                                         wOcc = (whiteOccupied & tt) != 0 ? whiteOccupied ^ tt : whiteOccupied;
//                                         occ = bOcc | wOcc;
                                        myMoves.add(new Move(t, tt, bPieces, wPieces, 0, pieceC == 5 ? 5 : -pieceC));
                                        m ^= tt;
                                    }
                                    l ^= t;
                                }
                                break;
                    case 5:     l = blackPieces[5];
                                int trailing = Long.numberOfTrailingZeros(l);
                                long m = kingMoves[trailing] & ~blackOccupied;
                                while (m != 0)
                                {
                                    long tt = Long.highestOneBit(m);
                                    long[] wPieces = new long[6];
                                    int pieceC = 5;
                                    for (int p = 0; p < 6; p++)
                                    {
                                        if ((whitePieces[p] & tt) != 0)
                                        {
                                            pieceC = p;
                                            wPieces[p] = whitePieces[p] ^ tt;
                                        }
                                        else
                                            wPieces[p] = whitePieces[p];
                                    }
                                    long[] bPieces = new long[6];
                                    for (int p = 0; p < 6; p++)
                                        bPieces[p] = blackPieces[p];
                                    bPieces[5] = bPieces[5] ^ l | tt;
//                                     bOcc = blackOccupied ^ l | tt;
//                                     wOcc = (whiteOccupied & tt) != 0 ? whiteOccupied ^ tt : whiteOccupied;
//                                     occ = bOcc | wOcc;
                                    myMoves.add(new Move(l, tt, bPieces, wPieces, 0, pieceC == 5 ? 5 : -pieceC));
                                    m ^= tt;
                                }
                                if (!bCastled && bK)
                                {
                                    if (bRA && (blackCastleSpacesA & ~occupied) == blackCastleSpacesA)
                                    {
                                        long[] bPieces = new long[6];
                                        for (int p = 0; p < 5; p++)
                                            bPieces[p] = blackPieces[p];
                                        bPieces[1] = bPieces[1] ^ squareBits[63] | squareBits[60];
                                        bPieces[5] = blackPieces[5] ^ squareBits[59] | squareBits[61];
//                                         bOcc = blackOccupied ^ squareBits[63] ^ squareBits[59] | squareBits[60] | squareBits[61];
//                                         occ = bOcc | whiteOccupied;
                                        Move move = new Move(l, squareBits[61], bPieces, whitePieces, 0, 2);
                                        move.castled = true;
                                        myMoves.add(move);
                                    }
                                    if (bRB && (blackCastleSpacesB & ~occupied) == blackCastleSpacesB)
                                    {
                                        long[] bPieces = new long[6];
                                        for (int p = 0; p < 5; p++)
                                            bPieces[p] = blackPieces[p];
                                        bPieces[1] = bPieces[1] ^ squareBits[56] | squareBits[58];
                                        bPieces[5] = blackPieces[5] ^ squareBits[59] | squareBits[57];
//                                         bOcc = blackOccupied ^ squareBits[56] ^ squareBits[59] | squareBits[58] | squareBits[57];
//                                         occ = bOcc | whiteOccupied;
                                        Move move = new Move(l, squareBits[57], bPieces, whitePieces, 0, 2);
                                        move.castled = true;
                                        myMoves.add(move);
                                    }
                                }
                }
                
            }
            Move myM = myMoves.first;
            while (myM != null)
            {
                long bOcc = myM.black[0];
                long wOcc = myM.white[0];
                for (int p = 1; p < 6; p++)
                {
                    bOcc |= myM.black[p];
                    wOcc |= myM.white[p];
                }
                long occ = bOcc | wOcc;
                if (!inCheck(myM.black[5], myM.white, occ, true))
                {
                    his = lookupBoard(myM.black, myM.white, depth);
                    if (his != null)
                        s = his;
                    else
                    {
                        s = calculate(depth + 1, occ, bOcc, wOcc, myM.black, myM.white, myM.castled ? true : bCastled, bRA, bRB, bK, wCastled, wRA, wRB, wK, myM.jumpPawn, score);
                        if (s != null)
                            storeBoard(myM.black, myM.white, s, depth);
                    }
                    if (s != null)
                    {
                        if (alpha != null && s > alpha)
                        {
                            bestFrom = myM.from;
                            bestTo = myM.to;
                            return null;
                        }
                        if (score == null || s > score || (s == score && Math.random() < .5))
                        {
                            score = s;
                            bestFrom = myM.from;
                            bestTo = myM.to;
                            if (depth == 0 && myM.promotion != -1)
                                promoPawn = myM.promotion;
                        }
                    }
                }
                myM = myM.next;
            }
            if (bestFrom == -1)
                if (inCheck(blackPieces[5], whitePieces, occupied, true))
                    score = -25000;
                else
                    score = -10000;
        }
        else
        {
            for (int i = 0; i < 6; i++)
            {
                switch (i)
                {
                    case 0:     l = whitePieces[0];
                                while (l != 0)
                                {
                                    long t = Long.highestOneBit(l);
                                    int trailing = Long.numberOfTrailingZeros(t);
                                    long m = whitePawnAttacks[trailing] & blackOccupied;
                                    while (m != 0)
                                    {
                                        long tt = Long.highestOneBit(m);
                                        long[] bPieces = new long[6];
                                        for (int p = 0; p < 6; p++)
                                            bPieces[p] = (blackPieces[p] & tt) != 0 ? blackPieces[p] ^ tt : blackPieces[p];
                                        long[] wPieces = new long[6];
                                        wPieces[0] = whitePieces[0] ^ t;
                                        for (int p = 1; p < 6; p++)
                                            wPieces[p] = whitePieces[p];
                                        if (Long.numberOfLeadingZeros(tt) < 9)
                                        {
                                            long[] wwPieces = new long[6];
                                            wwPieces[0] = wPieces[0];
                                            for (int p = 1; p < 5; p++)
                                            {
                                                for (int pp = 1; pp < 6; pp++)
                                                    if (pp == p)
                                                        wwPieces[pp] = whitePieces[pp] | tt;
                                                    else
                                                        wwPieces[pp] = whitePieces[pp];
//                                                 bOcc = blackOccupied ^ tt;
//                                                 wOcc = whiteOccupied ^ t | tt;
//                                                 occ = bOcc | wOcc;
                                                myMoves.add(new Move(t, tt, bPieces, wwPieces, 0, -6));
                                            }
                                        }
                                        else
                                        {
                                            wPieces[0] |= tt;
//                                             bOcc = blackOccupied ^ tt;
//                                             wOcc = whiteOccupied ^ t | tt;
//                                             occ = bOcc | wOcc;
                                            int pieceC = 0;
                                            for (pieceC = 0; pieceC < 5; pieceC++)
                                                if ((wPieces[pieceC] & tt) != 0)
                                                    break;
                                            myMoves.add(new Move(t, tt, bPieces, wPieces, 0, -pieceC));
                                        }
                                        m ^= tt;
                                    }
                                    m = whitePawnMoves[trailing];
                                    long tt = Long.lowestOneBit(m);
                                    if ((tt & ~occupied) != 0)
                                    {
                                        long[] wPieces = new long[6];
                                        wPieces[0] = whitePieces[0] ^ t;
                                        for (int p = 1; p < 6; p++)
                                            wPieces[p] = whitePieces[p];
                                        if (Long.numberOfLeadingZeros(tt) < 9)
                                        {
                                            long[] wwPieces = new long[6];
                                            wwPieces[0] = wPieces[0];
                                            for (int p = 1; p < 5; p++)
                                            {
                                                for (int pp = 1; pp < 6; pp++)
                                                    if (pp == p)
                                                        wwPieces[pp] = whitePieces[pp] | tt;
                                                    else
                                                        wwPieces[pp] = whitePieces[pp];
//                                                 wOcc = whiteOccupied ^ t | tt;
//                                                 occ = blackOccupied | wOcc;
                                                myMoves.add(new Move(t, tt, blackPieces, wwPieces, 0, 1));
                                            }
                                        }
                                        else
                                        {
                                            wPieces[0] |= tt;
//                                             wOcc = whiteOccupied ^ t | tt;
//                                             occ = blackOccupied | wOcc;
                                            myMoves.add(new Move(t, tt, blackPieces, wPieces, 0, 5));
                                            m ^= tt;
                                            long secondMove = Long.lowestOneBit(m);
                                            if ((secondMove & ~occupied) != 0)
                                            {
                                                wPieces[0] = whitePieces[0] ^ t | secondMove;
//                                                 wOcc = whiteOccupied ^ t | secondMove;
//                                                 occ = blackOccupied | wOcc;
                                                myMoves.add(new Move(t, tt, blackPieces, wPieces, tt, 5));
                                            }
                                        }
                                    }
                                    tt = whitePawnAttacks[trailing] & jumpPawn;
                                    if (tt != 0)
                                    {
                                        long pn = squareBits[Long.numberOfTrailingZeros(jumpPawn) + 8];
                                        long[] bPieces = new long[6];
                                        bPieces[0] = blackPieces[0] ^ pn;
                                        for (int p = 1; p < 6; p++)
                                            bPieces[p] = blackPieces[p];
                                        long[] wPieces = new long[6];
                                        for (int p = 1; p < 6; p++)
                                            wPieces[p] = whitePieces[p];
                                        wPieces[0] = whitePieces[0] ^ t | tt;
//                                         wOcc = whiteOccupied ^ t | tt;
//                                         bOcc = blackOccupied ^ pn;
//                                         occ = bOcc | wOcc;
                                        myMoves.add(new Move(t, tt, bPieces, wPieces, 0, 0));
                                    }
                                    l ^= t;
                                }
                                break;
                    case 4:
                    case 1:     l = whitePieces[i];
                                while (l != 0)
                                {
                                    long t = Long.highestOneBit(l);
                                    int trailing = Long.numberOfTrailingZeros(t);
                                    for (long lng : slideSouth.get(trailing))
                                        if ((occupied & lng) != 0)
                                        {
                                            if ((blackOccupied & lng) != 0)
                                            {
                                                long[] bPieces = new long[6];
                                                int pieceC = 0;
                                                for (int p = 0; p < 6; p++)
                                                {
                                                    if ((blackPieces[p] & lng) != 0)
                                                    {
                                                        pieceC = p;
                                                        bPieces[p] = blackPieces[p] ^ lng;
                                                    }
                                                    else
                                                        bPieces[p] = blackPieces[p];
                                                }
                                                long[] wPieces = new long[6];
                                                for (int p = 0; p < 6; p++)
                                                    if (p == i)
                                                        wPieces[p] = whitePieces[p] ^ t | lng;
                                                    else
                                                        wPieces[p] = whitePieces[p];
//                                                 wOcc = whiteOccupied ^ t | lng;
//                                                 bOcc = blackOccupied ^ lng;
//                                                 occ = bOcc | wOcc;
                                                myMoves.add(new Move(t, lng, bPieces, wPieces, 0, -pieceC));
                                            }
                                            break;
                                        }
                                        else
                                        {
                                            long[] wPieces = new long[6];
                                            for (int p = 0; p < 6; p++)
                                                if (p == i)
                                                    wPieces[p] = whitePieces[p] ^ t | lng;
                                                else
                                                    wPieces[p] = whitePieces[p];
//                                             wOcc = whiteOccupied ^ t | lng;
//                                             occ = blackOccupied | wOcc;
                                            myMoves.add(new Move(t, lng, blackPieces, wPieces, 0, 5));
                                        }
                                    for (long lng : slideEast.get(trailing))
                                        if ((occupied & lng) != 0)
                                        {
                                            if ((blackOccupied & lng) != 0)
                                            {
                                                long[] bPieces = new long[6];
                                                int pieceC = 0;
                                                for (int p = 0; p < 6; p++)
                                                {
                                                    if ((blackPieces[p] & lng) != 0)
                                                    {
                                                        pieceC = p;
                                                        bPieces[p] = blackPieces[p] ^ lng;
                                                    }
                                                    else
                                                        bPieces[p] = blackPieces[p];
                                                }
                                                long[] wPieces = new long[6];
                                                for (int p = 0; p < 6; p++)
                                                    if (p == i)
                                                        wPieces[p] = whitePieces[p] ^ t | lng;
                                                    else
                                                        wPieces[p] = whitePieces[p];
//                                                 wOcc = whiteOccupied ^ t | lng;
//                                                 bOcc = blackOccupied ^ lng;
//                                                 occ = bOcc | wOcc;
                                                myMoves.add(new Move(t, lng, bPieces, wPieces, 0, -pieceC));
                                            }
                                            break;
                                        }
                                        else
                                        {
                                            long[] wPieces = new long[6];
                                            for (int p = 0; p < 6; p++)
                                                if (p == i)
                                                    wPieces[p] = whitePieces[p] ^ t | lng;
                                                else
                                                    wPieces[p] = whitePieces[p];
//                                             wOcc = whiteOccupied ^ t | lng;
//                                             occ = blackOccupied | wOcc;
                                            myMoves.add(new Move(t, lng, blackPieces, wPieces, 0, 5));
                                        }
                                    for (long lng : slideWest.get(trailing))
                                        if ((occupied & lng) != 0)
                                        {
                                            if ((blackOccupied & lng) != 0)
                                            {
                                                long[] bPieces = new long[6];
                                                int pieceC = 0;
                                                for (int p = 0; p < 6; p++)
                                                {
                                                    if ((blackPieces[p] & lng) != 0)
                                                    {
                                                        pieceC = p;
                                                        bPieces[p] = blackPieces[p] ^ lng;
                                                    }
                                                    else
                                                        bPieces[p] = blackPieces[p];
                                                }
                                                long[] wPieces = new long[6];
                                                for (int p = 0; p < 6; p++)
                                                    if (p == i)
                                                        wPieces[p] = whitePieces[p] ^ t | lng;
                                                    else
                                                        wPieces[p] = whitePieces[p];
//                                                 wOcc = whiteOccupied ^ t | lng;
//                                                 bOcc = blackOccupied ^ lng;
//                                                 occ = bOcc | wOcc;
                                                myMoves.add(new Move(t, lng, bPieces, wPieces, 0, -pieceC));
                                            }
                                            break;
                                        }
                                        else
                                        {
                                            long[] wPieces = new long[6];
                                            for (int p = 0; p < 6; p++)
                                                if (p == i)
                                                    wPieces[p] = whitePieces[p] ^ t | lng;
                                                else
                                                    wPieces[p] = whitePieces[p];
//                                             wOcc = whiteOccupied ^ t | lng;
//                                             occ = blackOccupied | wOcc;
                                            myMoves.add(new Move(t, lng, blackPieces, wPieces, 0, 5));
                                        }
                                    for (long lng : slideNorth.get(trailing))
                                        if ((occupied & lng) != 0)
                                        {
                                            if ((blackOccupied & lng) != 0)
                                            {
                                                long[] bPieces = new long[6];
                                                int pieceC = 0;
                                                for (int p = 0; p < 6; p++)
                                                {
                                                    if ((blackPieces[p] & lng) != 0)
                                                    {
                                                        pieceC = p;
                                                        bPieces[p] = blackPieces[p] ^ lng;
                                                    }
                                                    else
                                                        bPieces[p] = blackPieces[p];
                                                }
                                                long[] wPieces = new long[6];
                                                for (int p = 0; p < 6; p++)
                                                    if (p == i)
                                                        wPieces[p] = whitePieces[p] ^ t | lng;
                                                    else
                                                        wPieces[p] = whitePieces[p];
//                                                 wOcc = whiteOccupied ^ t | lng;
//                                                 bOcc = blackOccupied ^ lng;
//                                                 occ = bOcc | wOcc;
                                                myMoves.add(new Move(t, lng, bPieces, wPieces, 0, -pieceC));
                                            }
                                            break;
                                        }
                                        else
                                        {
                                            long[] wPieces = new long[6];
                                            for (int p = 0; p < 6; p++)
                                                if (p == i)
                                                    wPieces[p] = whitePieces[p] ^ t | lng;
                                                else
                                                    wPieces[p] = whitePieces[p];
//                                             wOcc = whiteOccupied ^ t | lng;
//                                             occ = blackOccupied | wOcc;
                                            myMoves.add(new Move(t, lng, blackPieces, wPieces, 0, 5));
                                        }
                                    l ^= t;
                                }
                                if (i == 1)
                                    break;
                    case 3:     l = whitePieces[i];
                                while (l != 0)
                                {
                                    long t = Long.highestOneBit(l);
                                    int trailing = Long.numberOfTrailingZeros(t);
                                    for (long lng : slideNW.get(trailing))
                                        if ((occupied & lng) != 0)
                                        {
                                            if ((blackOccupied & lng) != 0)
                                            {
                                                long[] bPieces = new long[6];
                                                int pieceC = 0;
                                                for (int p = 0; p < 6; p++)
                                                {
                                                    if ((blackPieces[p] & lng) != 0)
                                                    {
                                                        pieceC = p;
                                                        bPieces[p] = blackPieces[p] ^ lng;
                                                    }
                                                    else
                                                        bPieces[p] = blackPieces[p];
                                                }
                                                long[] wPieces = new long[6];
                                                for (int p = 0; p < 6; p++)
                                                    if (p == i)
                                                        wPieces[p] = whitePieces[p] ^ t | lng;
                                                    else
                                                        wPieces[p] = whitePieces[p];
//                                                 wOcc = whiteOccupied ^ t | lng;
//                                                 bOcc = blackOccupied ^ lng;
//                                                 occ = bOcc | wOcc;
                                                myMoves.add(new Move(t, lng, bPieces, wPieces, 0, -pieceC));
                                            }
                                            break;
                                        }
                                        else
                                        {
                                            long[] wPieces = new long[6];
                                            for (int p = 0; p < 6; p++)
                                                if (p == i)
                                                    wPieces[p] = whitePieces[p] ^ t | lng;
                                                else
                                                    wPieces[p] = whitePieces[p];
//                                             wOcc = whiteOccupied ^ t | lng;
//                                             occ = blackOccupied | wOcc;
                                            myMoves.add(new Move(t, lng, blackPieces, wPieces, 0, 5));
                                        }
                                    for (long lng : slideNE.get(trailing))
                                        if ((occupied & lng) != 0)
                                        {
                                            if ((blackOccupied & lng) != 0)
                                            {
                                                long[] bPieces = new long[6];
                                                int pieceC = 0;
                                                for (int p = 0; p < 6; p++)
                                                {
                                                    if ((blackPieces[p] & lng) != 0)
                                                    {
                                                        pieceC = p;
                                                        bPieces[p] = blackPieces[p] ^ lng;
                                                    }
                                                    else
                                                        bPieces[p] = blackPieces[p];
                                                }
                                                long[] wPieces = new long[6];
                                                for (int p = 0; p < 6; p++)
                                                    if (p == i)
                                                        wPieces[p] = whitePieces[p] ^ t | lng;
                                                    else
                                                        wPieces[p] = whitePieces[p];
//                                                 wOcc = whiteOccupied ^ t | lng;
//                                                 bOcc = blackOccupied ^ lng;
//                                                 occ = bOcc | wOcc;
                                                myMoves.add(new Move(t, lng, bPieces, wPieces, 0, -pieceC));
                                            }
                                            break;
                                        }
                                        else
                                        {
                                            long[] wPieces = new long[6];
                                            for (int p = 0; p < 6; p++)
                                                if (p == i)
                                                    wPieces[p] = whitePieces[p] ^ t | lng;
                                                else
                                                    wPieces[p] = whitePieces[p];
//                                             wOcc = whiteOccupied ^ t | lng;
//                                             occ = blackOccupied | wOcc;
                                            myMoves.add(new Move(t, lng, blackPieces, wPieces, 0, 5));
                                        }
                                    for (long lng : slideSW.get(trailing))
                                        if ((occupied & lng) != 0)
                                        {
                                            if ((blackOccupied & lng) != 0)
                                            {
                                                long[] bPieces = new long[6];
                                                int pieceC = 0;
                                                for (int p = 0; p < 6; p++)
                                                {
                                                    if ((blackPieces[p] & lng) != 0)
                                                    {
                                                        pieceC = p;
                                                        bPieces[p] = blackPieces[p] ^ lng;
                                                    }
                                                    else
                                                        bPieces[p] = blackPieces[p];
                                                }
                                                long[] wPieces = new long[6];
                                                for (int p = 0; p < 6; p++)
                                                    if (p == i)
                                                        wPieces[p] = whitePieces[p] ^ t | lng;
                                                    else
                                                        wPieces[p] = whitePieces[p];
//                                                 wOcc = whiteOccupied ^ t | lng;
//                                                 bOcc = blackOccupied ^ lng;
//                                                 occ = bOcc | wOcc;
                                                myMoves.add(new Move(t, lng, bPieces, wPieces, 0, -pieceC));
                                            }
                                            break;
                                        }
                                        else
                                        {
                                            long[] wPieces = new long[6];
                                            for (int p = 0; p < 6; p++)
                                                if (p == i)
                                                    wPieces[p] = whitePieces[p] ^ t | lng;
                                                else
                                                    wPieces[p] = whitePieces[p];
//                                             wOcc = whiteOccupied ^ t | lng;
//                                             occ = blackOccupied | wOcc;
                                            myMoves.add(new Move(t, lng, blackPieces, wPieces, 0, 5));
                                        }
                                    for (long lng : slideSE.get(trailing))
                                        if ((occupied & lng) != 0)
                                        {
                                            if ((blackOccupied & lng) != 0)
                                            {
                                                long[] bPieces = new long[6];
                                                int pieceC = 0;
                                                for (int p = 0; p < 6; p++)
                                                {
                                                    if ((blackPieces[p] & lng) != 0)
                                                    {
                                                        pieceC = p;
                                                        bPieces[p] = blackPieces[p] ^ lng;
                                                    }
                                                    else
                                                        bPieces[p] = blackPieces[p];
                                                }
                                                long[] wPieces = new long[6];
                                                for (int p = 0; p < 6; p++)
                                                    if (p == i)
                                                        wPieces[p] = whitePieces[p] ^ t | lng;
                                                    else
                                                        wPieces[p] = whitePieces[p];
//                                                 wOcc = whiteOccupied ^ t | lng;
//                                                 bOcc = blackOccupied ^ lng;
//                                                 occ = bOcc | wOcc;
                                                myMoves.add(new Move(t, lng, bPieces, wPieces, 0, -pieceC));
                                            }
                                            break;
                                        }
                                        else
                                        {
                                            long[] wPieces = new long[6];
                                            for (int p = 0; p < 6; p++)
                                                if (p == i)
                                                    wPieces[p] = whitePieces[p] ^ t | lng;
                                                else
                                                    wPieces[p] = whitePieces[p];
//                                             wOcc = whiteOccupied ^ t | lng;
//                                             occ = blackOccupied | wOcc;
                                            myMoves.add(new Move(t, lng, blackPieces, wPieces, 0, 5));
                                        }
                                    l ^= t;
                                }
                                break;
                    case 2:     l = whitePieces[2];
                                while (l != 0)
                                {
                                    long t = Long.highestOneBit(l);
                                    int trailing = Long.numberOfTrailingZeros(t);
                                    long m = knightMoves[trailing] & ~whiteOccupied;
                                    while (m != 0)
                                    {
                                        long tt = Long.highestOneBit(m);
                                        long[] bPieces = new long[6];
                                        int pieceC = 5;
                                        for (int p = 0; p < 6; p++)
                                        {
                                            if ((blackPieces[p] & tt) != 0)
                                            {
                                                pieceC = p;
                                                bPieces[p] = blackPieces[p] ^ tt;
                                            }
                                            else
                                                bPieces[p] = blackPieces[p];
                                        }
                                        long[] wPieces = new long[6];
                                        for (int p = 0; p < 6; p++)
                                            wPieces[p] = whitePieces[p];
                                        wPieces[2] = wPieces[2] ^ t | tt;
//                                         wOcc = whiteOccupied ^ t | tt;
//                                         bOcc = (blackOccupied & tt) != 0 ? blackOccupied ^ tt : blackOccupied;
//                                         occ = bOcc | wOcc;
                                        myMoves.add(new Move(t, tt, bPieces, wPieces, 0, pieceC == 5 ? 5 : -pieceC));
                                        m ^= tt;
                                    }
                                    l ^= t;
                                }
                                break;
                    case 5:     l = whitePieces[5];
                                int trailing = Long.numberOfTrailingZeros(l);
                                long m = kingMoves[trailing] & ~whiteOccupied;
                                while (m != 0)
                                {
                                    long tt = Long.highestOneBit(m);
                                    long[] bPieces = new long[6];
                                    int pieceC = 5;
                                    for (int p = 0; p < 6; p++)
                                    {
                                        if ((blackPieces[p] & tt) != 0)
                                        {
                                            pieceC = p;
                                            bPieces[p] = blackPieces[p] ^ tt;
                                        }
                                        else
                                            bPieces[p] = blackPieces[p];
                                    }
                                    long[] wPieces = new long[6];
                                    for (int p = 0; p < 6; p++)
                                        wPieces[p] = whitePieces[p];
                                    wPieces[5] = wPieces[5] ^ l | tt;
//                                     wOcc = whiteOccupied ^ l | tt;
//                                     bOcc = (blackOccupied & tt) != 0 ? blackOccupied ^ tt : blackOccupied;
//                                     occ = bOcc | wOcc;
                                    myMoves.add(new Move(l, tt, bPieces, wPieces, 0, pieceC == 5 ? 5 : -pieceC));
                                    m ^= tt;
                                }
                                if (!wCastled && wK)
                                {
                                    if (wRA && (whiteCastleSpacesA & ~occupied) == whiteCastleSpacesA) // 0
                                    {
                                        long[] wPieces = new long[6];
                                        for (int p = 0; p < 5; p++)
                                            wPieces[p] = whitePieces[p];
                                        wPieces[1] = wPieces[1] ^ squareBits[7] | squareBits[4];
                                        wPieces[5] = whitePieces[5] ^ squareBits[3] | squareBits[5];
//                                         wOcc = whiteOccupied ^ squareBits[7] ^ squareBits[3] | squareBits[4] | squareBits[5];
//                                         occ = blackOccupied | wOcc;
                                        Move move = new Move(l, squareBits[5], blackPieces, wPieces, 0, 2);
                                        move.castled = true;
                                        myMoves.add(move);
                                    }
                                    if (wRB && (whiteCastleSpacesB & ~occupied) == whiteCastleSpacesB)
                                    {
                                        long[] wPieces = new long[6];
                                        for (int p = 0; p < 5; p++)
                                            wPieces[p] = whitePieces[p];
                                        wPieces[1] = wPieces[1] ^ squareBits[0] | squareBits[2];
                                        wPieces[5] = whitePieces[5] ^ squareBits[3] | squareBits[1];
//                                         wOcc = whiteOccupied ^ squareBits[0] ^ squareBits[3] | squareBits[2] | squareBits[1];
//                                         occ = blackOccupied | wOcc;
                                        Move move = new Move(l, squareBits[1], blackPieces, wPieces, 0, 2);
                                        move.castled = true;
                                        myMoves.add(move);
                                    }
                                }
                }
            }
            Move myM = myMoves.first;
            while (myM != null)
            {
                long bOcc = myM.black[0];
                long wOcc = myM.white[0];
                for (int p = 1; p < 6; p++)
                {
                    bOcc |= myM.black[p];
                    wOcc |= myM.white[p];
                }
                long occ = bOcc | wOcc;
                if (!inCheck(myM.white[5], myM.black, occ, false))
                {
                    his = lookupBoard(myM.black, myM.white, depth);
                    if (his != null)
                        s = his;
                    else
                    {
                        s = calculate(depth + 1, occ, bOcc, wOcc, myM.black, myM.white, bCastled, bRA, bRB, bK, myM.castled ? true : wCastled, wRA, wRB, wK, myM.jumpPawn, score);
                        if (s != null)
                            storeBoard(myM.black, myM.white, s, depth);
                    }
                    if (s != null)
                    {
                        if (alpha != null && s < alpha)
                        {
                            bestFrom = myM.from;
                            bestTo = myM.to;
                            return null;
                        }
                        if (score == null || s < score || (s == score && Math.random() < .5))
                        {
                            score = s;
                            bestFrom = myM.from;
                            bestTo = myM.to;
                        }
                    }
                }
                myM = myM.next;
            }
            if (bestFrom == -1)
                if (inCheck(whitePieces[5], blackPieces, occupied, false))
                    score = 25000;
                else
                    score = 10000;
        }
        if (depth == 0)
        {
            bestFrm = bestFrom;
            best2 = bestTo;
        }
        return score;
    }
    private boolean inCheck(long king, long[] pieces, long occupied, boolean blacksTurn)
    {
        int index = Long.numberOfTrailingZeros(king);
        // Knights
        if ((knightMoves[index] & pieces[2]) != 0)
            return true;
        // Pawns
        if ((pieces[0] & (blacksTurn ? blackPawnAttacks[index] : whitePawnAttacks[index])) != 0)
            return true;
        // King
        if ((pieces[5] & kingMoves[index]) != 0)
            return true;
        // Rooks & Queens
        for (Long l : slideSouth.get(index))
            if ((occupied & l) != 0)
            {
                if ((pieces[1] & l) != 0 || (pieces[4] & l) != 0)
                    return true;
                break;
            }
        for (Long l : slideEast.get(index))
            if ((occupied & l) != 0)
            {
                if ((pieces[1] & l) != 0 || (pieces[4] & l) != 0)
                    return true;
                break;
            }
        for (Long l : slideWest.get(index))
            if ((occupied & l) != 0)
            {
                if ((pieces[1] & l) != 0 || (pieces[4] & l) != 0)
                    return true;
                break;
            }
        for (Long l : slideNorth.get(index))
            if ((occupied & l) != 0)
            {
                if ((pieces[1] & l) != 0 || (pieces[4] & l) != 0)
                    return true;
                break;
            }
        // Bishops & Queens
        for (Long l : slideNW.get(index))
            if ((occupied & l) != 0)
            {
                if ((pieces[3] & l) != 0 || (pieces[4] & l) != 0)
                    return true;
                break;
            }
        for (Long l : slideNE.get(index))
            if ((occupied & l) != 0)
            {
                if ((pieces[3] & l) != 0 || (pieces[4] & l) != 0)
                    return true;
                break;
            }
        for (Long l : slideSW.get(index))
            if ((occupied & l) != 0)
            {
                if ((pieces[3] & l) != 0 || (pieces[4] & l) != 0)
                    return true;
                break;
            }
        for (Long l : slideSE.get(index))
            if ((occupied & l) != 0)
            {
                if ((pieces[3] & l) != 0 || (pieces[4] & l) != 0)
                    return true;
                break;
            }
        return false;
    }
    
    // Store a 1 in each square
    static long[] squareBits = new long[64];
    
    // Store move possibilities for pieces for each square of board
    static long[] rookMoves = new long[64];
    static long[] knightMoves = new long[64];
    static long[] bishopMoves = new long[64];
    static long[] queenMoves = new long[64];
    static long[] kingMoves = new long[64];
    static long[] blackPawnMoves = new long[64];
    static long[] blackPawnAttacks = new long[64];
    static long[] whitePawnMoves = new long[64];
    static long[] whitePawnAttacks = new long[64];
    
    // Store spaces that must be empty in order for a castle to take place
    static long blackCastleSpacesA;
    static long blackCastleSpacesB;
    static long whiteCastleSpacesA;
    static long whiteCastleSpacesB;
    
    static ArrayList<ArrayList<Long>> slideNorth = new ArrayList<ArrayList<Long>>();
    static ArrayList<ArrayList<Long>> slideEast = new ArrayList<ArrayList<Long>>();
    static ArrayList<ArrayList<Long>> slideWest = new ArrayList<ArrayList<Long>>();
    static ArrayList<ArrayList<Long>> slideSouth = new ArrayList<ArrayList<Long>>();
    
    static ArrayList<ArrayList<Long>> slideNW = new ArrayList<ArrayList<Long>>();
    static ArrayList<ArrayList<Long>> slideNE = new ArrayList<ArrayList<Long>>();
    static ArrayList<ArrayList<Long>> slideSE = new ArrayList<ArrayList<Long>>();
    static ArrayList<ArrayList<Long>> slideSW = new ArrayList<ArrayList<Long>>();
    
    static long middleOfBoard = 0;
    static long[] areaAroundKing = new long[64];
    
    private static int[][] HashKeyComponents;
    private static int[][] HashLockComponents;
    
    static void initialize()
    {
        for (int i = 0; i < 64; i++)
            squareBits[i] = (1L << i);
        for (int i = 0; i < 131072; i++)
            table[i] = new TableEntry();
        for (int y = 2; y < 6; y++)
            for (int x = 2; x < 6; x++)
                middleOfBoard |= squareBits[63 - y * 8 - x];
        for (int y = 0; y < 8; y++)
            for (int x = 0; x < 8; x++)
            {
                for (int i = -2; i < 3; i++)
                    for (int k = -2; k < 3; k++)
                        if (y + i > -1 && y + i < 8 && x + k > -1 && x + k < 8)
                            areaAroundKing[63 - y * 8 - x] |= squareBits[63 - (y + i) * 8 - (x + k)];
                slideNorth.add(0, new ArrayList<Long>());
                for (int i = y - 1; i > -1; i--)
                    slideNorth.get(0).add(squareBits[63 - i * 8 - x]);
                slideEast.add(0, new ArrayList<Long>());
                for (int i = x + 1; i < 8; i++)
                    slideEast.get(0).add(squareBits[63 - y * 8 - i]);
                slideWest.add(0, new ArrayList<Long>());
                for (int i = x - 1; i > -1; i--)
                    slideWest.get(0).add(squareBits[63 - y * 8 - i]);
                slideSouth.add(0, new ArrayList<Long>());
                for (int i = y + 1; i < 8; i++)
                    slideSouth.get(0).add(squareBits[63 - i * 8 - x]);
                
                slideNW.add(0, new ArrayList<Long>());
                for (int i = 1; x - i > -1 && y - i > -1; i++)
                    slideNW.get(0).add(squareBits[63 - (y - i) * 8 - (x - i)]);
                slideNE.add(0, new ArrayList<Long>());
                for (int i = 1; x + i < 8 && y - i > -1; i++)
                    slideNE.get(0).add(squareBits[63 - (y - i) * 8 - (x + i)]);
                slideSE.add(0, new ArrayList<Long>());
                for (int i = 1; x + i < 8 && y + i < 8; i++)
                    slideSE.get(0).add(squareBits[63 - (y + i) * 8 - (x + i)]);
                slideSW.add(0, new ArrayList<Long>());
                for (int i = 1; x - i > -1 && y + i < 8; i++)
                    slideSW.get(0).add(squareBits[63 - (y + i) * 8 - (x - i)]);
                
                long l = 0;
                for (int n = x - 1; n > -1; n--)
                    l |= (1L << ((7 - y) * 8 + 7 - n));
                for (int n = x + 1; n < 8; n++)
                    l |= (1L << ((7 - y) * 8 + 7 - n));
                for (int n = y - 1; n > -1; n--)
                    l |= (1L << ((7 - n) * 8 + 7 - x));
                for (int n = y + 1; n < 8; n++)
                    l |= (1L << ((7 - n) * 8 + 7 - x));
                rookMoves[63 - y * 8 - x] = l;
                l = 0;
                for (int n = -1; x + n > -1 && y + n > -1; n--)
                    l |= (1L << ((7 - (y + n)) * 8 + 7 - (x + n)));
                for (int n = 1; x + n < 8 && y - n > -1; n++)
                    l |= (1L << ((7 - (y - n)) * 8 + 7 - (x + n)));
                for (int n = 1; x + n < 8 && y + n < 8; n++)
                    l |= (1L << ((7 - (y + n)) * 8 + 7 - (x + n)));
                for (int n = 1; x - n > -1 && y + n < 8; n++)
                    l |= (1L << ((7 - (y + n)) * 8 + 7 - (x - n)));
                bishopMoves[63 - y * 8 - x] = l;
                queenMoves[63 - y * 8 - x] = l | rookMoves[63 - y * 8 - x];
                l = 0;
                for (int i = -2; i < 3; i += 4)
                    for (int k = -1; k < 2; k += 2)
                        if (y + k > -1 && y + k < 8 && x + i > -1 && x + i < 8)
                            l |= (1L << ((7 - (y + k)) * 8 + 7 - (x + i)));
                for (int i = -1; i < 2; i += 2)
                    for (int k = -2; k < 3; k += 4)
                        if (y + k > -1 && y + k < 8 && x + i > -1 && x + i < 8)
                            l |= (1L << ((7 - (y + k)) * 8 + 7 - (x + i)));
                knightMoves[63 - y * 8 - x] = l;
                l = 0;
                for (int i = -1; i < 2; i++)
                    for (int k = -1; k < 2; k++)
                        if (y + k > -1 && y + k < 8 && x + i > -1 && x + i < 8 && (k != 0 || i != 0))
                            l |= (1L << ((7 - (y + k)) * 8 + 7 - (x + i)));
                kingMoves[63 - y * 8 - x] = l;
                l = 0;
                if (y != 7)
                    l |= (1L << ((7 - (y + 1)) * 8 + 7 - x));
                if (y == 1)
                    l |= (1L << ((7 - (y + 2)) * 8 + 7 - x));
                blackPawnMoves[63 - y * 8 - x] = l;
                l = 0;
                if (y != 7)
                {
                    if (x > 0)
                        l |= (1L << ((7 - (y + 1)) * 8 + 7 - (x - 1)));
                    if (x < 7)
                        l |= (1L << ((7 - (y + 1)) * 8 + 7 - (x + 1)));
                }
                blackPawnAttacks[63 - y * 8 - x] = l;
                l = 0;
                if (y != 0)
                    l |= (1L << ((7 - (y - 1)) * 8 + 7 - x));
                if (y == 6)
                    l |= (1L << ((7 - (y - 2)) * 8 + 7 - x));
                whitePawnMoves[63 - y * 8 - x] = l;
                l = 0;
                if (y != 0)
                {
                    if (x > 0)
                        l |= (1L << ((7 - (y - 1)) * 8 + 7 - (x - 1)));
                    if (x < 7)
                        l |= (1L << ((7 - (y - 1)) * 8 + 7 - (x + 1)));
                }
                whitePawnAttacks[63 - y * 8 - x] = l;
            }
        blackCastleSpacesA = 0;
        for (int i = 1; i < 4; i++)
            blackCastleSpacesA |= (1L << (7 * 8 + 7 - i));
        blackCastleSpacesB = 0;
        for (int i = 5; i < 7; i++)
            blackCastleSpacesB |= (1L << (7 * 8 + 7 - i));
        whiteCastleSpacesA = 0;
        for (int i = 1; i < 4; i++)
            whiteCastleSpacesA |= (1L << (7 - i));
        whiteCastleSpacesB = 0;
        for (int i = 5; i < 7; i++)
            whiteCastleSpacesB |= (1L << (7 - i));
        
        HashKeyComponents = new int[12][64];
        HashLockComponents = new int[12][64];
        Random rnd = new Random();
        for(int i = 0; i < 12; i++)
        {
            for(int j = 0; j < 64; j++)
            {
                HashKeyComponents[i][j] = rnd.nextInt();
                HashLockComponents[i][j] = rnd.nextInt();
            }
        }
    }
    public static void printOut(long l)
    {
        String s = Long.toBinaryString(l);
        for (int i = 64 - s.length(); i > 0; i--)
            s = "0" + s;
        for (int i = 0; i < 8; i++)
            System.out.println(s.substring(i * 8, (i + 1) * 8));
        System.out.println("");
    }
    public Integer lookupBoard(long[] bPieces, long[] wPieces, int depth)
    {
        // Find the board's hash position in Table
        int key = Math.abs(getHashKey(bPieces, wPieces) % 131072);
        TableEntry entry = table[key];

        // If the entry is an empty placeholder, we don't have a match
        if (entry.depth == -1 || entry.depth > depth)
            return null;

        // Check for a hashing collision!
        if (entry.lock != getHashLock(bPieces, wPieces))
            return null;

        // Now, we know that we have a match!
        return entry.score;
    }
    public void storeBoard(long[] bPieces, long[] wPieces, int score, int depth)
    {
        int key = Math.abs(getHashKey(bPieces, wPieces) % 131072);

        // Would we erase a more useful (i.e., higher) position if we stored this
        // one?  If so, don't bother!
        if ((table[key].depth != -1) && (table[key].depth > depth) && (table[key].age >= depth + clock))
            return;

        // And now, do the actual work
        table[key].lock = getHashLock(bPieces, wPieces);
        table[key].score = score;
        table[key].depth = depth;
        table[key].age = depth + clock;
        return;
    }
    private static int getHashKey(long[] bPieces, long[] wPieces)
    {
        int hash = 0;
        for (int s = 0; s < 64; s++)
        {
            for (int i = 0; i < 6; i++)
                if ((bPieces[i] & squareBits[s]) != 0)
                    hash ^= HashKeyComponents[i][s];
            for (int i = 0; i < 6; i++)
                if ((wPieces[i] & squareBits[s]) != 0)
                    hash ^= HashKeyComponents[6 + i][s];
        }
        return hash;
    }
    private static int getHashLock(long[] bPieces, long[] wPieces)
    {
        int hash = 0;
        for (int s = 0; s < 64; s++)
        {
            for (int i = 0; i < 6; i++)
                if ((bPieces[i] & squareBits[s]) != 0)
                    hash ^= HashLockComponents[i][s];
            for (int i = 0; i < 6; i++)
                if ((wPieces[i] & squareBits[s]) != 0)
                    hash ^= HashLockComponents[6 + i][s];
        }
        return hash;
    }
}