import greenfoot.*;  // (World, Actor, GreenfootImage, Greenfoot and MouseInfo)
/**
 * Write a description of class Board here.
 * 
 * @author Taylor Born
 * @version September 2010
 */
import java.util.*;
import java.awt.Color;
public class Board  extends Actor
{
    private int turn;
    private Square activeSquare;
    private ArrayList<Piece> pieces;
    private ArrayList<ArrayList<Square>> grid;
    private ArrayList<ArrayList<Square>> tiles;
    private GreenfootImage black = new GreenfootImage("black.png");
    private GreenfootImage white = new GreenfootImage("white.png");
    private GreenfootImage active = new GreenfootImage("active.png");
    private GreenfootImage from = new GreenfootImage("from.png");
    private Square cSquare;
    private Point lastMouse = new Point(-25, -25);
    private Point bKing;
    private Point wKing;
    private boolean inCheck;
    private boolean inCheckMate;
    private boolean inStaleMate;
    private boolean wCastled;
    private boolean wCastleRookA;
    private boolean wCastleRookB;
    private boolean wCastleKing;
    private boolean bCastled;
    private boolean bCastleRookA;
    private boolean bCastleRookB;
    private boolean bCastleKing;
    private Point dragOffset;
    private boolean dragging;
    private Square promote;
    private Square jumpPawn;
    private GreenfootSound checkSnd = new GreenfootSound("check.wav");
    private GreenfootSound checkMateSnd = new GreenfootSound("checkMate.wav");
    private GreenfootSound staleMateSnd = new GreenfootSound("staleMate.wav");
    private GreenfootSound castleKSnd = new GreenfootSound("castleKing.wav");
    private GreenfootSound castleQSnd = new GreenfootSound("castleQueen.wav");
    private GreenfootSound pawnSnd = new GreenfootSound("pawn.wav");
    private GreenfootSound rookSnd = new GreenfootSound("rook.wav");
    private GreenfootSound knightSnd = new GreenfootSound("knight.wav");
    private GreenfootSound bishopSnd = new GreenfootSound("bishop.wav");
    private GreenfootSound queenSnd = new GreenfootSound("queen.wav");
    private GreenfootSound kingSnd = new GreenfootSound("king.wav");
    private GreenfootSound toSnd = new GreenfootSound("to.wav");
    private GreenfootSound takesSnd = new GreenfootSound("takes.wav");
    private ArrayList<GreenfootSound> columns = new ArrayList<GreenfootSound>();
    private ArrayList<GreenfootSound> rows = new ArrayList<GreenfootSound>();
    private GreenfootSound activeSound = checkSnd;
    private boolean markers = true;
    private String keysPressed = "";
    private String voice;
    private int soundStatus = 1;
    private int fade = 255;
    private String mess = "New Human vs. Human";
    private boolean computer = false;
    private AI myAI;
    private int waiting;
    private GreenfootImage thinking = new GreenfootImage("think.jpg");
    private long lastTime;
    private double player1Time;
    private double player2Time;
    private boolean started;
    private static final int CODE_ID = 28;
    private boolean displayedCode = false;
    
    public Board()
    {
        newGame();
        for (int i = 0; i < 8; i++)
            columns.add(new GreenfootSound((char)(97 + i) + ".wav"));
        for (int i = 8; i > 0; i--)
            rows.add(new GreenfootSound(i + ".wav"));
    }
    public void switchToAI()
    {
        computer = true;
    }
    public void newGame()
    {
        turn = 1;
        started = false;
        player1Time = 0;
        player2Time = 0;
        grid = new ArrayList<ArrayList<Square>>();
        for (int i = 0; i < 8; i++)
        {
            grid.add(new ArrayList<Square>());
            for (int k = 0; k < 8; k++)
                grid.get(i).add(new Square());
        }
        pieces = new ArrayList<Piece>();
        for (int i = 1; i < 7; i += 5)
            for (int k = 0; k < 8; k++)
            {
                pieces.add(new Piece(0, i == 1 ? 2 : 1));
                grid.get(i).get(k).setPiece(pieces.get(pieces.size() - 1));
            }
        for (int i = 0; i < 8; i += 7)
            for (int k = 0; k < 8; k += 7)
            {
                pieces.add(new Piece(1, i == 0 ? 2 : 1));
                grid.get(i).get(k).setPiece(pieces.get(pieces.size() - 1));
            }
        for (int i = 0; i < 8; i += 7)
            for (int k = 1; k < 7; k += 5)
            {
                pieces.add(new Piece(2, i == 0 ? 2 : 1));
                grid.get(i).get(k).setPiece(pieces.get(pieces.size() - 1));
            }
        for (int i = 0; i < 8; i += 7)
            for (int k = 2; k < 6; k += 3)
            {
                pieces.add(new Piece(3, i == 0 ? 2 : 1));
                grid.get(i).get(k).setPiece(pieces.get(pieces.size() - 1));
            }
        pieces.add(new Piece(4, 2));
        grid.get(0).get(3).setPiece(pieces.get(pieces.size() - 1));
        pieces.add(new Piece(4, 1));
        grid.get(7).get(3).setPiece(pieces.get(pieces.size() - 1));
        pieces.add(new Piece(5, 2));
        grid.get(0).get(4).setPiece(pieces.get(pieces.size() - 1));
        bKing = new Point(4, 0);
        pieces.add(new Piece(5, 1));
        grid.get(7).get(4).setPiece(pieces.get(pieces.size() - 1));
        wKing = new Point(4, 7);
        for (int i = 0; i < 8; i++)
            for (int k = 0; k < 8; k++)
                grid.get(i).get(k).givePos(new Point(k, i));
        tiles = convert();
        wCastled = false;
        wCastleRookA = true;
        wCastleRookB = true;
        wCastleKing = true;
        bCastled = false;
        bCastleRookA = true;
        bCastleRookB = true;
        bCastleKing = true;
        promote = null;
        activeSquare = null;
        dragging = false;
        inCheck = false;
        inCheckMate = false;
        inStaleMate = false;
        voice = "";
        System.out.println("-new game-");
    }
    public void paint()
    {
        MouseInfo mouse = Greenfoot.getMouseInfo();
        cSquare = null;
        boolean marker = true;
        GreenfootImage pic = new GreenfootImage(400, 300);
        for (int y = 0; y < tiles.size(); y++)
        {
            int xOffSet = 13;
            if (y < tiles.size() / 2)
                xOffSet += 24 * (tiles.size() / 2 - y);
            else
                xOffSet += 24 * (y - tiles.size() / 2);
            for (int x = 0; x < tiles.get(y).size(); x++)
            {
                pic.setColor(Color.WHITE);
                if (y >= tiles.size() / 2 && markers)
                {
                    if (x == 0)
                        pic.drawString((char)(97 + y - tiles.size() / 2) + "", xOffSet + x * 48 + 17 - 22, 34 + 37 + y * 14 - 15 - 21 + 30);
                    if (x == tiles.get(y).size() - 1)
                        pic.drawString(8 - (y - tiles.size() / 2) + "", xOffSet + x * 48 + 17 - 22 + 40, 34 + 37 + y * 14 - 15 - 21 + 30);
                }
                pic.drawImage(y % 2 == 0 ? white : black, xOffSet + x * 48 + 17 - 22, 34 + 37 + y * 14 - 15 - 21);
                if (promote == null && !inCheckMate && !inStaleMate && (computer ? turn == 1 : true))
                {
                    boolean getFirst = Greenfoot.isKeyDown("f");
                    if (tiles.get(y).get(x).getPiece() != null && (getFirst ? activeSquare == null : true) && tiles.get(y).get(x).getPiece().getPlayer() == turn && Greenfoot.mousePressed(this) && mouse.getX() > xOffSet + x * 48 + 17 - 22 + 13 && mouse.getX() < xOffSet + x * 48 + 17 - 22 + 13 + 23 && mouse.getY() > 34 + 37 + y * 14 - 15 - 21 - 35 && mouse.getY() < 34 + 37 + y * 14 - 15 - 21 - 35 + 54 && !tiles.get(y).get(x).getPiece().getPic().getColorAt(mouse.getX() - (xOffSet + x * 48 + 17 - 22 + 13), mouse.getY() - (34 + 37 + y * 14 - 15 - 21 - 35)).equals(black.getColorAt(0, 0)))
                    {
                        activeSquare = tiles.get(y).get(x);
                        dragOffset = new Point(mouse.getX() - (xOffSet + x * 48 + 17 - 22 + 13), mouse.getY() - (34 + 37 + y * 14 - 15 - 21 - 35));
                        dragging = true;
                    }
                    if (activeSquare == tiles.get(y).get(x))
                        pic.drawImage(from, xOffSet + x * 48 + 17 - 22, 34 + 37 + y * 14 - 15 - 21);
                    if (marker && dragging && lastMouse.getX() - dragOffset.getX() + 12 > xOffSet + x * 48 + 17 - 22 && lastMouse.getX() - dragOffset.getX() + 12 < xOffSet + x * 48 + 17 - 22 + 49 && lastMouse.getY() - dragOffset.getY() + 54 > 34 + 37 + y * 14 - 15 - 21 && lastMouse.getY() - dragOffset.getY() + 54 < 34 + 37 + y * 14 - 15 - 21 + 29 && !black.getColorAt(lastMouse.getX() - dragOffset.getX() + 12 - (xOffSet + x * 48 + 17 - 22), lastMouse.getY() - dragOffset.getY() + 54 - (34 + 37 + y * 14 - 15 - 21)).equals(black.getColorAt(0, 0)))
                    {
                        pic.drawImage(active, xOffSet + x * 48 + 17 - 22, 34 + 37 + y * 14 - 15 - 21);
                        marker = false;
                    }
                    if (dragging && Greenfoot.mouseDragEnded(this) && mouse.getX() - dragOffset.getX() + 12 > xOffSet + x * 48 + 17 - 22 && mouse.getX() - dragOffset.getX() + 12 < xOffSet + x * 48 + 17 - 22 + 49 && mouse.getY() - dragOffset.getY() + 54 > 34 + 37 + y * 14 - 15 - 21 && mouse.getY() - dragOffset.getY() + 54 < 34 + 37 + y * 14 - 15 - 21 + 29 && !black.getColorAt(mouse.getX() - dragOffset.getX() + 12 - (xOffSet + x * 48 + 17 - 22), mouse.getY() - dragOffset.getY() + 54 - (34 + 37 + y * 14 - 15 - 21)).equals(black.getColorAt(0, 0)))
                    {
                        cSquare = tiles.get(y).get(x);
                        move();
                        activeSquare = null;
                        dragging = false;
                    }
                }
                if (tiles.get(y).get(x).getPiece() != null && activeSquare != tiles.get(y).get(x))
                    pic.drawImage(tiles.get(y).get(x).getPiece().getPic(), xOffSet + x * 48 + 17 - 22 + 13, 34 + 37 + y * 14 - 15 - 21 - 35);
            }
        }
        if (promote != null)
        {
            pic.drawString("Choose a piece:", 20, 20);
            pic.drawImage(new GreenfootImage(turn + "1.png"), 20, 20);
            pic.drawImage(new GreenfootImage(turn + "2.png"), 43, 20);
            pic.drawImage(new GreenfootImage(turn + "3.png"), 66, 20);
            pic.drawImage(new GreenfootImage(turn + "4.png"), 89, 20);
            int selection = -1;
            if (Greenfoot.mousePressed(this))
                if (mouse.getY() > 20 && mouse.getY() < 74 && mouse.getX() > 20 && mouse.getX() < 112)
                    selection = (mouse.getX() - 20) / 23;
            if (selection != -1)
            {
                promote.getPiece().promote(selection + 1);
                if (turn == 1)
                    turn = 2;
                else
                    turn = 1;
                inCheck = check(null, null, null);
                if (inCheck)
                {
                    inCheckMate = checkMate();
                    if (inCheckMate)
                        checkMateSnd.play();
                    else
                        checkSnd.play();
                }
                else if (staleMate())
                    inStaleMate = true;
                if (!inCheckMate && !inStaleMate && computer && turn == 2)
                    waiting = 2;
                promote = null;
            }
        }
        if (waiting > 0)
            pic.drawImage(thinking, 305, -35);
        if (dragging)
            pic.drawImage(activeSquare.getPiece().getPic(), lastMouse.getX() - dragOffset.getX(), lastMouse.getY() - dragOffset.getY());
        if (inStaleMate)
        {
            pic.drawString((turn == 1 ? "White" : "Black") + " is in", 15, 242);
            pic.drawString("stalemate", 15, 257);
        }
        else if (inCheckMate)
        {
            pic.drawString((turn == 1 ? "White" : "Black") + " is in", 15, 242);
            pic.drawString("checkmate", 15, 257);
            if (!displayedCode)
            {
                displayedCode = true;
                if (CodeGenerator.haveUser())
                {
                    getWorld().removeObjects(getWorld().getObjects(CodeDisplay.class));
                    getWorld().addObject(new CodeDisplay(CODE_ID), getWorld().getWidth() / 2, getWorld().getHeight() - 6);
                }
            }
        }
        else
        {
            pic.drawString((turn == 1 ? "White" : "Black") + "'s turn", 15, 242);
            if (inCheck)
                pic.drawString("You are in check.", 15, 257);
        }
        if (fade > 0)
        {
            GreenfootImage message = new GreenfootImage(400, 20);
            message.setColor(Color.WHITE);
            message.drawString(mess, 10, 15);
            message.setTransparency(fade);
            pic.drawImage(message, 0, 280);
            fade -= 1;
        }
        pic.drawString((int)(player1Time / 60) + ":" + ((int)(player1Time % 60) < 10 ? "0" : "") + (int)(player1Time % 60), 155, 20);
        pic.drawString((int)(player2Time / 60) + ":" + ((int)(player2Time % 60) < 10 ? "0" : "") + (int)(player2Time % 60), 220, 20);
        setImage(pic);
        if ((Greenfoot.mouseDragEnded(this) || Greenfoot.mouseClicked(this)) && dragging)
        {
            activeSquare = null;
            dragging = false;
        }
    }
    public void act() 
    {
        if (waiting > 0)
        {
            Greenfoot.setSpeed(100);
            waiting--;
            if (waiting == 0)
            {
                myAI = new AI(grid, bCastled, bCastleRookA, bCastleRookB, bCastleKing, wCastled, wCastleRookA, wCastleRookB, wCastleKing, jumpPawn == null ? null : jumpPawn.getPos());
                getWorld().addObject(myAI, 375, 25);
            }
        }
        if (started)
            if (turn == 1)
                player1Time += (System.currentTimeMillis() - lastTime) / 1000d;
            else
                player2Time += (System.currentTimeMillis() - lastTime) / 1000d;
        lastTime = System.currentTimeMillis();
        MouseInfo mouse = Greenfoot.getMouseInfo();
        if (Greenfoot.mouseMoved(this) || Greenfoot.mouseDragged(this))
            lastMouse = new Point(mouse.getX(), mouse.getY());
        paint();
        if (Greenfoot.isKeyDown("1"))
        {
            if (!keysPressed.contains(" 1 "))
            {
                newGame();
                computer = false;
                mess = "New Human vs. Human";
                fade = 255;
                keysPressed += " 1 ";
            }
        }
        else
            keysPressed = keysPressed.replace(" 1 ", "");
        if (Greenfoot.isKeyDown("2"))
        {
            if (!keysPressed.contains(" 2 "))
            {
                newGame();
                computer = true;
                mess = "New Human vs. Computer";
                fade = 255;
                keysPressed += " 2 ";
            }
        }
        else
            keysPressed = keysPressed.replace(" 2 ", "");
        if (myAI != null && myAI.finished)
        {
            activeSquare = grid.get(myAI.getBestFrom().getY()).get(myAI.getBestFrom().getX());
            cSquare = grid.get(myAI.getBestTo().getY()).get(myAI.getBestTo().getX());
            move();
            getWorld().removeObject(myAI);
            myAI = null;
        }
        if (Greenfoot.isKeyDown("0"))
        {
            if (!keysPressed.contains(" 0 "))
            {
                markers = !markers;
                keysPressed += " 0 ";
            }
        }
        else
            keysPressed = keysPressed.replace(" 0 ", "");
        if (Greenfoot.isKeyDown("9"))
        {
            if (!keysPressed.contains(" 9 "))
            {
                soundStatus++;
                if (soundStatus == 4)
                    soundStatus = 1;
                fade = 255;
                mess = soundStatus == 1 ? "Sound On" : soundStatus == 2 ? "Only Check, Checkmate, and Stalemate Sounds" : "Sound Off";
                keysPressed += " 9 ";
            }
        }
        else
            keysPressed = keysPressed.replace(" 9 ", "");
        if (!voice.equals("") && !activeSound.isPlaying())
        {
            if (voice.startsWith("castleK"))
                activeSound = castleKSnd;
            else if (voice.startsWith("castleQ"))
                activeSound = castleQSnd;
            else if (voice.startsWith("0"))
                activeSound = pawnSnd;
            else if (voice.startsWith("1"))
                activeSound = rookSnd;
            else if (voice.startsWith("2"))
                activeSound = knightSnd;
            else if (voice.startsWith("3"))
                activeSound = bishopSnd;
            else if (voice.startsWith("4"))
                activeSound = queenSnd;
            else if (voice.startsWith("5"))
                activeSound = kingSnd;
            else if (voice.startsWith("to"))
                activeSound = toSnd;
            else if (voice.startsWith("takes"))
                activeSound = takesSnd;
            else if (voice.startsWith("x"))
            {
                voice = voice.substring(1);
                activeSound = columns.get(Integer.parseInt(voice.substring(0, voice.indexOf(","))));
            }
            else if (voice.startsWith("y"))
            {
                voice = voice.substring(1);
                activeSound = rows.get(Integer.parseInt(voice.substring(0, voice.indexOf(","))));
            }
            else if (voice.startsWith("stalemate"))
                activeSound = staleMateSnd;
            else if (voice.startsWith("checkmate"))
                activeSound = checkMateSnd;
            else if (voice.startsWith("check"))
                activeSound = checkSnd;
            activeSound.play();
            voice = voice.substring(voice.indexOf(",") + 1);
        }
    }
    public void move()
    {
        if (activeSquare == cSquare)
            activeSquare = null;
        if (cSquare.getPiece() == null || (cSquare.getPiece() != null && cSquare.getPiece().getPlayer() != turn))
        {
            boolean jump = false;
            String castle = "";
            boolean able = false;
            boolean flag = false;
            Piece p = activeSquare.getPiece();
            Point curPos = activeSquare.getPos();
            Point desPos = cSquare.getPos();
            switch (p.getType())
            {
                case 0:     if (cSquare.getPiece() == null)
                            {
                                if ((desPos.getX() == curPos.getX() && ((p.getPlayer() == 2 && curPos.getY() == 1 && desPos.getY() == 3 && grid.get(curPos.getY() + 1).get(curPos.getX()).getPiece() == null) || (p.getPlayer() == 1 && curPos.getY() == 6 && desPos.getY() == 4 && grid.get(curPos.getY() - 1).get(curPos.getX()).getPiece() == null) || (p.getPlayer() == 2 ? desPos.getY() == curPos.getY() + 1 : desPos.getY() == curPos.getY() - 1))) || (cSquare == jumpPawn && Math.abs(curPos.getX() - desPos.getX()) == 1 && desPos.getY() == curPos.getY() + (turn == 1 ? -1 : 1)))
                                {
                                    able = true;
                                    jump = Math.abs(curPos.getY() - desPos.getY()) == 2;
                                }
                            }
                            else if ((p.getPlayer() == 2 ? desPos.getY() == curPos.getY() + 1 : desPos.getY() == curPos.getY() - 1) && Math.abs(desPos.getX() - curPos.getX()) == 1)
                                able = true;
                break;
                case 1:     if (curPos.getX() == desPos.getX() || curPos.getY() == desPos.getY())
                            {
                                for (int i = (desPos.getX() < curPos.getX() || desPos.getY() < curPos.getY() ? -1 : 1); (curPos.getX() == desPos.getX() ? (desPos.getY() < curPos.getY() ? curPos.getY() + i > desPos.getY() : curPos.getY() + i < desPos.getY()) : (desPos.getX() < curPos.getX() ? curPos.getX() + i > desPos.getX() : curPos.getX() + i < desPos.getX())); i += (desPos.getX() < curPos.getX() || desPos.getY() < curPos.getY() ? -1 : 1))
                                    if (grid.get(curPos.getY() + (desPos.getY() != curPos.getY() ? i : 0)).get(curPos.getX() + (desPos.getX() != curPos.getX() ? i : 0)).getPiece() != null)
                                        flag = true;
                            }
                            else
                                flag = true;
                            if (!flag)
                                able = true;
                break;
                case 2:     if ((Math.abs(curPos.getX() - desPos.getX()) == 2 && Math.abs(curPos.getY() - desPos.getY()) == 1) || (Math.abs(curPos.getX() - desPos.getX()) == 1 && Math.abs(curPos.getY() - desPos.getY()) == 2))
                                able = true;
                break;
                case 3:     if (Math.abs(curPos.getX() - desPos.getX()) == Math.abs(curPos.getY() - desPos.getY()))
                            {
                                for (int i = 1; i < Math.abs(curPos.getX() - desPos.getX()); i++)
                                    if (grid.get(curPos.getY() + (desPos.getY() < curPos.getY() ? -i : i)).get(curPos.getX() + (desPos.getX() < curPos.getX() ? -i : i)).getPiece() != null)
                                        flag = true;
                            }
                            else
                                flag = true;
                            if (!flag)
                                able = true;
                break;
                case 4:     if (curPos.getX() == desPos.getX() || curPos.getY() == desPos.getY())
                            {
                                for (int i = (desPos.getX() < curPos.getX() || desPos.getY() < curPos.getY() ? -1 : 1); (curPos.getX() == desPos.getX() ? (desPos.getY() < curPos.getY() ? curPos.getY() + i > desPos.getY() : curPos.getY() + i < desPos.getY()) : (desPos.getX() < curPos.getX() ? curPos.getX() + i > desPos.getX() : curPos.getX() + i < desPos.getX())); i += (desPos.getX() < curPos.getX() || desPos.getY() < curPos.getY() ? -1 : 1))
                                    if (grid.get(curPos.getY() + (desPos.getY() != curPos.getY() ? i : 0)).get(curPos.getX() + (desPos.getX() != curPos.getX() ? i : 0)).getPiece() != null)
                                        flag = true;
                            }
                            else if (Math.abs(curPos.getX() - desPos.getX()) == Math.abs(curPos.getY() - desPos.getY()))
                            {
                                for (int i = 1; i < Math.abs(curPos.getX() - desPos.getX()); i++)
                                    if (grid.get(curPos.getY() + (desPos.getY() < curPos.getY() ? -i : i)).get(curPos.getX() + (desPos.getX() < curPos.getX() ? -i : i)).getPiece() != null)
                                        flag = true;
                            }
                            else
                                flag = true;
                            if (!flag)
                                able = true;
                break;
                case 5:     if (Math.abs(curPos.getX() - desPos.getX()) < 2 && Math.abs(curPos.getY() - desPos.getY()) < 2)
                                able = true;
                            else if (curPos.getY() == desPos.getY())
                                if (turn == 1 && wCastleKing)
                                {
                                    if (curPos.getX() - 2 == desPos.getX() && wCastleRookA)
                                    {
                                        for (int i = 1; i < 4; i++)
                                            if (grid.get(7).get(4 - i).getPiece() != null)
                                                flag = true;
                                        if (!flag)
                                            castle = "wL";
                                    }
                                    else if (curPos.getX() + 2 == desPos.getX() && wCastleRookB)
                                    {
                                        for (int i = 1; i < 3; i++)
                                            if (grid.get(7).get(4 + i).getPiece() != null)
                                                flag = true;
                                        if (!flag)
                                            castle = "wR";
                                    }
                                }
                                else if (bCastleKing)
                                {
                                    if (curPos.getX() - 2 == desPos.getX() && bCastleRookA)
                                    {
                                        for (int i = 1; i < 4; i++)
                                            if (grid.get(0).get(4 - i).getPiece() != null)
                                                flag = true;
                                        if (!flag)
                                            castle = "bL";
                                    }
                                    else if (curPos.getX() + 2 == desPos.getX() && bCastleRookB)
                                    {
                                        for (int i = 1; i < 3; i++)
                                            if (grid.get(0).get(4 + i).getPiece() != null)
                                                flag = true;
                                        if (!flag)
                                            castle = "bR";
                                    }
                                }
                            if (!castle.equals(""))
                                able = true;
            }
            if (able)
            {
                Piece old = cSquare.getPiece();
                Piece attacker = activeSquare.getPiece();
                if (p.getType() == 5)
                    (p.getPlayer() == 1 ? wKing : bKing).set(cSquare.getPos().getX(), cSquare.getPos().getY());
                if (p.getType() == 0 && cSquare == jumpPawn)
                {
                    pieces.remove(grid.get(curPos.getY()).get(jumpPawn.getPos().getX()).getPiece());
                    grid.get(curPos.getY()).get(jumpPawn.getPos().getX()).setPiece(null);
                }
                pieces.remove(cSquare.getPiece());
                cSquare.setPiece(activeSquare.getPiece());
                activeSquare.setPiece(null);
                if (castle.equals("wL"))
                {
                    grid.get(7).get(3).setPiece(grid.get(7).get(0).getPiece());
                    grid.get(7).get(0).setPiece(null);
                }
                else if (castle.equals("wR"))
                {
                    grid.get(7).get(5).setPiece(grid.get(7).get(7).getPiece());
                    grid.get(7).get(7).setPiece(null);
                }
                else if (castle.equals("bL"))
                {
                    grid.get(0).get(3).setPiece(grid.get(0).get(0).getPiece());
                    grid.get(0).get(0).setPiece(null);
                }
                else if (castle.equals("bR"))
                {
                    grid.get(0).get(5).setPiece(grid.get(0).get(7).getPiece());
                    grid.get(0).get(7).setPiece(null);
                }
                if (check(null, null, null))
                {
                    cSquare.setPiece(old);
                    activeSquare.setPiece(attacker);
                    pieces.add(old);
                    if (p.getType() == 5)
                        (p.getPlayer() == 1 ? wKing : bKing).set(activeSquare.getPos().getX(), activeSquare.getPos().getY());
                    if (p.getType() == 0 && cSquare == jumpPawn)
                    {
                        pieces.add(new Piece(0, (turn == 1 ? 2 : 1)));
                        grid.get(curPos.getY()).get(jumpPawn.getPos().getX()).setPiece(pieces.get(pieces.size() - 1));
                    }
                    if (castle.equals("wL"))
                    {
                        grid.get(7).get(0).setPiece(grid.get(7).get(3).getPiece());
                        grid.get(7).get(3).setPiece(null);
                    }
                    else if (castle.equals("wR"))
                    {
                        grid.get(7).get(7).setPiece(grid.get(7).get(5).getPiece());
                        grid.get(7).get(5).setPiece(null);
                    }
                    else if (castle.equals("bL"))
                    {
                        grid.get(0).get(0).setPiece(grid.get(0).get(3).getPiece());
                        grid.get(0).get(3).setPiece(null);
                    }
                    else if (castle.equals("bR"))
                    {
                        grid.get(0).get(7).setPiece(grid.get(0).get(5).getPiece());
                        grid.get(0).get(5).setPiece(null);
                    }
                }
                else
                {
                    if (started)
                        if (turn == 1)
                            player1Time += (System.currentTimeMillis() - lastTime) / 1000d;
                        else
                            player2Time += (System.currentTimeMillis() - lastTime) / 1000d;
                    lastTime = System.currentTimeMillis();
                    if (castle.startsWith("w"))
                        wCastled = true;
                    else if (castle.startsWith("b"))
                        bCastled = true;
                    if (activeSquare.getPos().equals(new Point(0, 7)) || cSquare.getPos().equals(new Point(0, 7)))
                        wCastleRookA = false;
                    if (activeSquare.getPos().equals(new Point(7, 7)) || cSquare.getPos().equals(new Point(7, 7)))
                        wCastleRookB = false;
                    if (activeSquare.getPos().equals(new Point(0, 0)) || cSquare.getPos().equals(new Point(0, 0)))
                        bCastleRookA = false;
                    if (cSquare.getPos().equals(new Point(7, 0)) || cSquare.getPos().equals(new Point(7, 0)))
                        bCastleRookB = false;
                    if (p.getType() == 5)
                        if (p.getPlayer() == 1)
                            wCastleKing = false;
                        else
                            bCastleKing = false;
                    if (p.getType() == 0)
                        if (desPos.getY() == 0 || desPos.getY() == 7)
                            if (turn == 2 && computer)
                                cSquare.getPiece().promote(myAI.getPromotion());
                            else
                                promote = cSquare;
                    p.setLoc(cSquare.getPos());
                    if (jump)
                        jumpPawn = grid.get(desPos.getY() + (turn == 1 ? 1 : -1)).get(desPos.getX());
                    else
                        jumpPawn = null;
                    if (promote == null)
                    {
                        if (turn == 1)
                            turn = 2;
                        else
                            turn = 1;
                        inCheck = check(null, null, null);
                        if (inCheck)
                            inCheckMate = checkMate();
                        else if (staleMate())
                            inStaleMate = true;
                    }
                    if (inCheckMate || inStaleMate)
                        started = false;
                    else
                        started = true;
                    voice = "";
                    activeSound.stop();
                    if (soundStatus != 3)
                    {
                        if (soundStatus != 2)
                        {
                            if (castle.contains("L"))
                                voice += "castleQ,";
                            else if (castle.contains("R"))
                                voice += "castleK,";
                            else
                            {
                                voice += p.getType() + ",";
                                if (old != null)
                                    voice += "takes,";
                                else
                                    voice += "to,";
                                voice += "x" + cSquare.getPos().getX() + ",";
                                voice += "y" + cSquare.getPos().getY() + ",";
                            }
                        }
                        if (inStaleMate)
                            voice += "stalemate,";
                        else if (inCheckMate)
                            voice += "checkmate,";
                        else if (inCheck)
                            voice += "check,";
                    }
                    System.out.println((char)(97 + activeSquare.getPos().getX()) + "" + (8 - activeSquare.getPos().getY()) + "-" + (char)(97 + cSquare.getPos().getX()) + "" + (8 - cSquare.getPos().getY()));
                    if (turn == 2 && computer && !inCheckMate && !inStaleMate)
                        waiting = 2;
                }
                activeSquare = null;
            }
        }
    }
    public boolean check(Point kingN, ArrayList<Piece> ignore, ArrayList<Piece> add)
    {
        Point start = kingN != null ? kingN : turn == 1 ? wKing : bKing;
        if (kingN != null)
        {
            if (grid.get(kingN.getY()).get(kingN.getX()).getPiece() != null)
                ignore.add(grid.get(kingN.getY()).get(kingN.getX()).getPiece());
            Point pt = turn == 1 ? wKing : bKing;
            add.add(new Piece(grid.get(pt.getY()).get(pt.getX()).getPiece(), start));
            ignore.add(grid.get(pt.getY()).get(pt.getX()).getPiece());
        }
        if (ignore != null)
        {
            for (Piece p : ignore)
                if (p != null)
                    grid.get(p.getLoc().getY()).get(p.getLoc().getX()).setPiece(null);
            for (Piece p : add)
                grid.get(p.getLoc().getY()).get(p.getLoc().getX()).setPiece(p);
        }
        for (int i = 1; start.getX() - i > -1; i++)
            if (grid.get(start.getY()).get(start.getX() - i).getPiece() != null)
            {
                if (grid.get(start.getY()).get(start.getX() - i).getPiece().getPlayer() != turn && "14".contains(grid.get(start.getY()).get(start.getX() - i).getPiece().getType() + ""))
                {
                    changeBack(ignore, add);
                    return true;
                }
                break;
            }
        for (int i = 1; start.getX() + i < 8; i++)
            if (grid.get(start.getY()).get(start.getX() + i).getPiece() != null)
            {
                if (grid.get(start.getY()).get(start.getX() + i).getPiece().getPlayer() != turn && "14".contains(grid.get(start.getY()).get(start.getX() + i).getPiece().getType() + ""))
                {
                    changeBack(ignore, add);
                    return true;
                }
                break;
            }
        for (int i = 1; start.getY() - i > -1; i++)
            if (grid.get(start.getY() - i).get(start.getX()).getPiece() != null)
            {
                if (grid.get(start.getY() - i).get(start.getX()).getPiece().getPlayer() != turn && "14".contains(grid.get(start.getY() - i).get(start.getX()).getPiece().getType() + ""))
                {
                    changeBack(ignore, add);
                    return true;
                }
                break;
            }
        for (int i = 1; start.getY() + i < 8; i++)
            if (grid.get(start.getY() + i).get(start.getX()).getPiece() != null)
            {
                if (grid.get(start.getY() + i).get(start.getX()).getPiece().getPlayer() != turn && "14".contains(grid.get(start.getY() + i).get(start.getX()).getPiece().getType() + ""))
                {
                    changeBack(ignore, add);
                    return true;
                }
                break;
            }
        for (int i = 1; start.getX() - i > -1 && start.getY() - i > -1; i++)
            if (grid.get(start.getY() - i).get(start.getX() - i).getPiece() != null)
            {
                if (grid.get(start.getY() - i).get(start.getX() - i).getPiece().getPlayer() != turn && "34".contains(grid.get(start.getY() - i).get(start.getX() - i).getPiece().getType() + ""))
                {
                    changeBack(ignore, add);
                    return true;
                }
                break;
            }
        for (int i = 1; start.getX() + i < 8 && start.getY() + i < 8; i++)
            if (grid.get(start.getY() + i).get(start.getX() + i).getPiece() != null)
            {
                if (grid.get(start.getY() + i).get(start.getX() + i).getPiece().getPlayer() != turn && "34".contains(grid.get(start.getY() + i).get(start.getX() + i).getPiece().getType() + ""))
                {
                    changeBack(ignore, add);
                    return true;
                }
                break;
            }
        for (int i = 1; start.getX() + i < 8 && start.getY() - i > -1; i++)
            if (grid.get(start.getY() - i).get(start.getX() + i).getPiece() != null)
            {
                if (grid.get(start.getY() - i).get(start.getX() + i).getPiece().getPlayer() != turn && "34".contains(grid.get(start.getY() - i).get(start.getX() + i).getPiece().getType() + ""))
                {
                    changeBack(ignore, add);
                    return true;
                }
                break;
            }
        for (int i = 1; start.getX() - i > -1 && start.getY() + i < 8; i++)
            if (grid.get(start.getY() + i).get(start.getX() - i).getPiece() != null)
            {
                if (grid.get(start.getY() + i).get(start.getX() - i).getPiece().getPlayer() != turn && "34".contains(grid.get(start.getY() + i).get(start.getX() - i).getPiece().getType() + ""))
                {
                    changeBack(ignore, add);
                    return true;
                }
                break;
            }
        for (int i = -2; i < 3; i += 4)
            for (int k = -1; k < 2; k += 2)
                if (start.getX() + i > -1 && start.getX() + i < 8 && start.getY() + k > -1 && start.getY() + k < 8 && grid.get(start.getY() + k).get(start.getX() + i).getPiece() != null && grid.get(start.getY() + k).get(start.getX() + i).getPiece().getPlayer() != turn && grid.get(start.getY() + k).get(start.getX() + i).getPiece().getType() == 2)
                {
                    changeBack(ignore, add);
                    return true;
                }
        for (int i = -1; i < 2; i += 2)
            for (int k = -2; k < 3; k += 4)
                if (start.getX() + i > -1 && start.getX() + i < 8 && start.getY() + k > -1 && start.getY() + k < 8 && grid.get(start.getY() + k).get(start.getX() + i).getPiece() != null && grid.get(start.getY() + k).get(start.getX() + i).getPiece().getPlayer() != turn && grid.get(start.getY() + k).get(start.getX() + i).getPiece().getType() == 2)
                {
                    changeBack(ignore, add);
                    return true;
                }
        for (int i = -1; i < 2; i += 2)
            if (start.getY() + (turn == 1 ? -1 : 1) < 8 && start.getY() + (turn == 1 ? -1 : 1) > -1 && start.getX() + i < 8 && start.getX() + i > -1 && grid.get(start.getY() + (turn == 1 ? -1 : 1)).get(start.getX() + i).getPiece() != null && grid.get(start.getY() + (turn == 1 ? -1 : 1)).get(start.getX() + i).getPiece().getPlayer() != turn && grid.get(start.getY() + (turn == 1 ? -1 : 1)).get(start.getX() + i).getPiece().getType() == 0)
            {
                changeBack(ignore, add);
                return true;
            }
        if (Math.abs((kingN != null ? kingN.getX() - (turn == 1 ? bKing.getX() : wKing.getX()) : wKing.getX() - bKing.getX())) < 2 && Math.abs((kingN != null ? kingN.getY() - (turn == 1 ? bKing.getY() : wKing.getY()) : wKing.getY() - bKing.getY())) < 2)
        {
            changeBack(ignore, add);
            return true;
        }
        changeBack(ignore, add);
        return false;
    }
    public void changeBack(ArrayList<Piece> ignore, ArrayList<Piece> add)
    {
        if (add != null)
            for (Piece p : add)
                grid.get(p.getLoc().getY()).get(p.getLoc().getX()).setPiece(null);
        if (ignore != null)
            for (Piece p : ignore)
                if (p != null)
                    grid.get(p.getLoc().getY()).get(p.getLoc().getX()).setPiece(p);
    }
    public boolean checkMate()
    {
        Point start = turn == 1 ? wKing : bKing;
        for (int i = start.getY() - 1; i < start.getY() + 2; i++)
            if (i > -1 && i < 8)
                for (int k = start.getX() - 1; k < start.getX() + 2; k++)
                    if (k > -1 && k < 8 && (i != start.getY() || k != start.getX()) && (grid.get(i).get(k).getPiece() == null ? true : grid.get(i).get(k).getPiece().getPlayer() != turn) && !check(new Point(k, i), new ArrayList<Piece>(), new ArrayList<Piece>()))
                        return false;
        if (turn == 1 && wCastleKing)
        {
            if (wCastleRookA)
            {
                boolean flag = false;
                for (int i = 1; i < 4; i++)
                    if (grid.get(7).get(4 - i).getPiece() != null)
                        flag = true;
                if (!flag)
                {
                    ArrayList<Piece> ignore = new ArrayList<Piece>();
                    ArrayList<Piece> add = new ArrayList<Piece>();
                    add.add(new Piece(grid.get(0).get(7).getPiece(), new Point(3, 7)));
                    ignore.add(grid.get(0).get(7).getPiece());
                    if (!check(new Point(2, 7), ignore, add))
                        return false;
                }
            }
            if (wCastleRookB)
            {
                boolean flag = false;
                for (int i = 1; i < 3; i++)
                    if (grid.get(7).get(4 + i).getPiece() != null)
                        flag = true;
                if (!flag)
                {
                    ArrayList<Piece> ignore = new ArrayList<Piece>();
                    ArrayList<Piece> add = new ArrayList<Piece>();
                    add.add(new Piece(grid.get(7).get(7).getPiece(), new Point(5, 7)));
                    ignore.add(grid.get(7).get(7).getPiece());
                    if (!check(new Point(6, 7), ignore, add))
                        return false;
                }
            }
        }
        else if (turn == 2 && bCastleKing)
        {
            if (bCastleRookA)
            {
                boolean flag = false;
                for (int i = 1; i < 4; i++)
                    if (grid.get(0).get(4 - i).getPiece() != null)
                        flag = true;
                if (!flag)
                {
                    ArrayList<Piece> ignore = new ArrayList<Piece>();
                    ArrayList<Piece> add = new ArrayList<Piece>();
                    add.add(new Piece(grid.get(0).get(0).getPiece(), new Point(3, 0)));
                    ignore.add(grid.get(0).get(0).getPiece());
                    if (!check(new Point(2, 0), ignore, add))
                        return false;
                }
            }
            if (bCastleRookB)
            {
                boolean flag = false;
                for (int i = 1; i < 3; i++)
                    if (grid.get(0).get(4 + i).getPiece() != null)
                        flag = true;
                if (!flag)
                {
                    ArrayList<Piece> ignore = new ArrayList<Piece>();
                    ArrayList<Piece> add = new ArrayList<Piece>();
                    add.add(new Piece(grid.get(7).get(0).getPiece(), new Point(5, 0)));
                    ignore.add(grid.get(7).get(0).getPiece());
                    if (!check(new Point(6, 0), ignore, add))
                        return false;
                }
            }
        }
        ArrayList<ArrayList<Point>> paths = new ArrayList<ArrayList<Point>>();
        ArrayList<Point> points = new ArrayList<Point>();
        for (int i = 1; start.getX() - i > -1; i++)
        {
            points.add(new Point(start.getX() - i, start.getY()));
            if (grid.get(start.getY()).get(start.getX() - i).getPiece() != null)
            {
                if (grid.get(start.getY()).get(start.getX() - i).getPiece().getPlayer() != turn && "14".contains(grid.get(start.getY()).get(start.getX() - i).getPiece().getType() + ""))
                    paths.add(points);
                break;
            }
        }
        points = new ArrayList<Point>();
        for (int i = 1; start.getX() + i < 8; i++)
        {
            points.add(new Point(start.getX() + i, start.getY()));
            if (grid.get(start.getY()).get(start.getX() + i).getPiece() != null)
            {
                if (grid.get(start.getY()).get(start.getX() + i).getPiece().getPlayer() != turn && "14".contains(grid.get(start.getY()).get(start.getX() + i).getPiece().getType() + ""))
                    paths.add(points);
                break;
            }
        }
        points = new ArrayList<Point>();
        for (int i = 1; start.getY() - i > -1; i++)
        {
            points.add(new Point(start.getX(), start.getY() - i));
            if (grid.get(start.getY() - i).get(start.getX()).getPiece() != null)
            {
                if (grid.get(start.getY() - i).get(start.getX()).getPiece().getPlayer() != turn && "14".contains(grid.get(start.getY() - i).get(start.getX()).getPiece().getType() + ""))
                    paths.add(points);
                break;
            }
        }
        points = new ArrayList<Point>();
        for (int i = 1; start.getY() + i < 8; i++)
        {
            points.add(new Point(start.getX(), start.getY() + i));
            if (grid.get(start.getY() + i).get(start.getX()).getPiece() != null)
            {
                if (grid.get(start.getY() + i).get(start.getX()).getPiece().getPlayer() != turn && "14".contains(grid.get(start.getY() + i).get(start.getX()).getPiece().getType() + ""))
                    paths.add(points);
                break;
            }
        }
        points = new ArrayList<Point>();
        for (int i = 1; start.getX() - i > -1 && start.getY() - i > -1; i++)
        {
            points.add(new Point(start.getX() - i, start.getY() - i));
            if (grid.get(start.getY() - i).get(start.getX() - i).getPiece() != null)
            {
                if (grid.get(start.getY() - i).get(start.getX() - i).getPiece().getPlayer() != turn && "34".contains(grid.get(start.getY() - i).get(start.getX() - i).getPiece().getType() + ""))
                    paths.add(points);
                break;
            }
        }
        points = new ArrayList<Point>();
        for (int i = 1; start.getX() + i < 8 && start.getY() + i < 8; i++)
        {
            points.add(new Point(start.getX() + i, start.getY() + i));
            if (grid.get(start.getY() + i).get(start.getX() + i).getPiece() != null)
            {
                if (grid.get(start.getY() + i).get(start.getX() + i).getPiece().getPlayer() != turn && "34".contains(grid.get(start.getY() + i).get(start.getX() + i).getPiece().getType() + ""))
                    paths.add(points);
                break;
            }
        }
        points = new ArrayList<Point>();
        for (int i = 1; start.getX() + i < 8 && start.getY() - i > -1; i++)
        {
            points.add(new Point(start.getX() + i, start.getY() - i));
            if (grid.get(start.getY() - i).get(start.getX() + i).getPiece() != null)
            {
                if (grid.get(start.getY() - i).get(start.getX() + i).getPiece().getPlayer() != turn && "34".contains(grid.get(start.getY() - i).get(start.getX() + i).getPiece().getType() + ""))
                    paths.add(points);
                break;
            }
        }
        points = new ArrayList<Point>();
        for (int i = 1; start.getX() - i > -1 && start.getY() + i < 8; i++)
        {
            points.add(new Point(start.getX() - i, start.getY() + i));
            if (grid.get(start.getY() + i).get(start.getX() - i).getPiece() != null)
            {
                if (grid.get(start.getY() + i).get(start.getX() - i).getPiece().getPlayer() != turn && "34".contains(grid.get(start.getY() + i).get(start.getX() - i).getPiece().getType() + ""))
                    paths.add(points);
                break;
            }
        }
        for (int i = -2; i < 3; i += 4)
            for (int k = -1; k < 2; k += 2)
                if (start.getX() + i > -1 && start.getX() + i < 8 && start.getY() + k > -1 && start.getY() + k < 8 && grid.get(start.getY() + k).get(start.getX() + i).getPiece() != null && grid.get(start.getY() + k).get(start.getX() + i).getPiece().getPlayer() != turn && grid.get(start.getY() + k).get(start.getX() + i).getPiece().getType() == 2)
                {
                    points = new ArrayList<Point>();
                    points.add(new Point(start.getX() + i, start.getY() + k));
                    paths.add(points);
                }
        for (int i = -1; i < 2; i += 2)
            for (int k = -2; k < 3; k += 4)
                if (start.getX() + i > -1 && start.getX() + i < 8 && start.getY() + k > -1 && start.getY() + k < 8 && grid.get(start.getY() + k).get(start.getX() + i).getPiece() != null && grid.get(start.getY() + k).get(start.getX() + i).getPiece().getPlayer() != turn && grid.get(start.getY() + k).get(start.getX() + i).getPiece().getType() == 2)
                {
                    points = new ArrayList<Point>();
                    points.add(new Point(start.getX() + i, start.getY() + k));
                    paths.add(points);
                }
        for (int i = -1; i < 2; i += 2)
            if (start.getY() + (turn == 1 ? -1 : 1) < 8 && start.getY() + (turn == 1 ? -1 : 1) > -1 && start.getX() + i < 8 && start.getX() + i > -1 && grid.get(start.getY() + (turn == 1 ? -1 : 1)).get(start.getX() + i).getPiece() != null && grid.get(start.getY() + (turn == 1 ? -1 : 1)).get(start.getX() + i).getPiece().getPlayer() != turn && grid.get(start.getY() + (turn == 1 ? -1 : 1)).get(start.getX() + i).getPiece().getType() == 0)
            {
                points = new ArrayList<Point>();
                points.add(new Point(start.getX() + i, start.getY() + (turn == 1 ? -1 : 1)));
                paths.add(points);
            }
        for (ArrayList<Point> path : paths)
        {
            for (int p = 0; p < path.size(); p++)
            {
                for (int p2 = 0; p2 < paths.size(); p2++)
                {
                    boolean flag = true;
                    for (Point aPoint : paths.get(p2))
                        if (path.get(p).equals(aPoint))
                            flag = false;
                    if (flag)
                    {
                        path.remove(path.get(p));
                        p--;
                        p2 = paths.size();
                    }
                }
            }
            if (path.size() == 0)
                return true;
        }
        if (paths.size() > 0)
            for (int p = 0; p < paths.get(0).size(); p++)
            {
                for (int i = 0; i < pieces.size(); i++)
                    if (pieces.get(i).getPlayer() == turn && canBlock(pieces.get(i), paths.get(0).get(p), p == paths.get(0).size() - 1))
                        return false;
            }
        return true;
    }
    public boolean canBlock(Piece p, Point pnt, boolean forTheKill)
    {
        Point curPos = p.getLoc();
        boolean able = false;
        boolean flag = false;
        switch (p.getType())
        {
            case 0:     if (grid.get(pnt.getY()).get(pnt.getX()).getPiece() == null)
                        {
                            if ((pnt.getX() == curPos.getX() && ((p.getPlayer() == 2 && curPos.getY() == 1 && pnt.getY() == 3 && !forTheKill && grid.get(curPos.getY() + 1).get(curPos.getX()).getPiece() == null) || (p.getPlayer() == 1 && curPos.getY() == 6 && pnt.getY() == 4 && !forTheKill && grid.get(curPos.getY() - 1).get(curPos.getX()).getPiece() == null) || (p.getPlayer() == 2 ? pnt.getY() == curPos.getY() + 1 : pnt.getY() == curPos.getY() - 1))))
                                able = true;
                        }
                        else if (((p.getPlayer() == 2 ? pnt.getY() == curPos.getY() + 1 : pnt.getY() == curPos.getY() - 1) && Math.abs(pnt.getX() - curPos.getX()) == 1) || (grid.get(pnt.getY() + (turn == 1 ? -1 : 1)).get(pnt.getX()) == jumpPawn && Math.abs(curPos.getX() - pnt.getX()) == 1 && pnt.getY() == curPos.getY()))
                            able = true;
            break;
            case 1:     if (curPos.getX() == pnt.getX() || curPos.getY() == pnt.getY())
                        {
                            for (int i = (pnt.getX() < curPos.getX() || pnt.getY() < curPos.getY() ? -1 : 1); (curPos.getX() == pnt.getX() ? (pnt.getY() < curPos.getY() ? curPos.getY() + i > pnt.getY() : curPos.getY() + i < pnt.getY()) : (pnt.getX() < curPos.getX() ? curPos.getX() + i > pnt.getX() : curPos.getX() + i < pnt.getX())); i += (pnt.getX() < curPos.getX() || pnt.getY() < curPos.getY() ? -1 : 1))
                                if (grid.get(curPos.getY() + (pnt.getY() != curPos.getY() ? i : 0)).get(curPos.getX() + (pnt.getX() != curPos.getX() ? i : 0)).getPiece() != null)
                                    flag = true;
                        }
                        else
                            flag = true;
                        if (!flag)
                            able = true;
            break;
            case 2:     if ((Math.abs(curPos.getX() - pnt.getX()) == 2 && Math.abs(curPos.getY() - pnt.getY()) == 1) || (Math.abs(curPos.getX() - pnt.getX()) == 1 && Math.abs(curPos.getY() - pnt.getY()) == 2))
                            able = true;
            break;
            case 3:     if (Math.abs(curPos.getX() - pnt.getX()) == Math.abs(curPos.getY() - pnt.getY()))
                        {
                            for (int i = 1; i < Math.abs(curPos.getX() - pnt.getX()); i++)
                                if (grid.get(curPos.getY() + (pnt.getY() < curPos.getY() ? -i : i)).get(curPos.getX() + (pnt.getX() < curPos.getX() ? -i : i)).getPiece() != null)
                                    flag = true;
                        }
                        else
                            flag = true;
                        if (!flag)
                            able = true;
            break;
            case 4:     if (curPos.getX() == pnt.getX() || curPos.getY() == pnt.getY())
                        {
                            for (int i = (pnt.getX() < curPos.getX() || pnt.getY() < curPos.getY() ? -1 : 1); (curPos.getX() == pnt.getX() ? (pnt.getY() < curPos.getY() ? curPos.getY() + i > pnt.getY() : curPos.getY() + i < pnt.getY()) : (pnt.getX() < curPos.getX() ? curPos.getX() + i > pnt.getX() : curPos.getX() + i < pnt.getX())); i += (pnt.getX() < curPos.getX() || pnt.getY() < curPos.getY() ? -1 : 1))
                                if (grid.get(curPos.getY() + (pnt.getY() != curPos.getY() ? i : 0)).get(curPos.getX() + (pnt.getX() != curPos.getX() ? i : 0)).getPiece() != null)
                                    flag = true;
                        }
                        else if (Math.abs(curPos.getX() - pnt.getX()) == Math.abs(curPos.getY() - pnt.getY()))
                        {
                            for (int i = 1; i < Math.abs(curPos.getX() - pnt.getX()); i++)
                                if (grid.get(curPos.getY() + (pnt.getY() < curPos.getY() ? -i : i)).get(curPos.getX() + (pnt.getX() < curPos.getX() ? -i : i)).getPiece() != null)
                                    flag = true;
                        }
                        else
                            flag = true;
                        if (!flag)
                            able = true;
        }
        if (able)
        {
            ArrayList<Piece> ignore = new ArrayList<Piece>();
            ArrayList<Piece> add = new ArrayList<Piece>();
            add.add(new Piece(p, pnt));
            if (grid.get(pnt.getY()).get(pnt.getX()).getPiece() != null)
                ignore.add(grid.get(pnt.getY()).get(pnt.getX()).getPiece());
            ignore.add(p);
            if (!check(null, ignore, add))
                return true;
        }
        return false;
    }
    public boolean staleMate()
    {
        for (Piece p : pieces)
        {
            if (p.getPlayer() == turn)
            {
                ArrayList<Piece> ignore;
                ArrayList<Piece> add;
                if (p.getType() == 0)
                {
                    ArrayList<Point> spots = new ArrayList<Point>();
                    if (grid.get(p.getLoc().getY() + (turn == 2 ? 1 : -1)).get(p.getLoc().getX()).getPiece() == null)
                    {
                        spots.add(new Point(p.getLoc().getX(), p.getLoc().getY() + (turn == 2 ? 1 : -1)));
                        if (p.getLoc().getY() == (turn == 1 ? 6 : 1) && grid.get(p.getLoc().getY() + (turn == 2 ? 2 : -2)).get(p.getLoc().getX()).getPiece() == null)
                            spots.add(new Point(p.getLoc().getX(), p.getLoc().getY() + (turn == 2 ? 2 : -2)));
                    }
                    ignore = new ArrayList<Piece>();
                    ignore.add(p);
                    for (Point pnt : spots)
                    {
                        add = new ArrayList<Piece>();
                        add.add(new Piece(p, pnt));
                        if (!check(null, ignore, add))
                            return false;
                    }
                    for (int i = -1; i < 2; i += 2)
                    {
                        if (p.getLoc().getX() + i > -1 && p.getLoc().getX() + i < 8)
                            if (grid.get(p.getLoc().getY() + (turn == 2 ? 1 : -1)).get(p.getLoc().getX() + i).getPiece() != null && grid.get(p.getLoc().getY() + (turn == 2 ? 1 : -1)).get(p.getLoc().getX() + i).getPiece().getPlayer() != turn)
                            {
                                ignore = new ArrayList<Piece>();
                                add = new ArrayList<Piece>();
                                ignore.add(p);
                                ignore.add(grid.get(p.getLoc().getY() + (turn == 2 ? 1 : -1)).get(p.getLoc().getX() + i).getPiece());
                                add.add(new Piece(p, new Point(p.getLoc().getX() + i, p.getLoc().getY() + (turn == 2 ? 1 : -1))));
                                if (!check(null, ignore, add))
                                    return false;
                            }
                            else if (grid.get(p.getLoc().getY() + (turn == 2 ? 1 : -1)).get(p.getLoc().getX() + i) == jumpPawn)
                            {
                                ignore = new ArrayList<Piece>();
                                add = new ArrayList<Piece>();
                                ignore.add(p);
                                ignore.add(grid.get(p.getLoc().getY()).get(p.getLoc().getX() + i).getPiece());
                                add.add(new Piece(p, new Point(p.getLoc().getX() + i, p.getLoc().getY() + (turn == 2 ? 1 : -1))));
                                if (!check(null, ignore, add))
                                    return false;
                            }
                    }
                }
                if (p.getType() == 1 || p.getType() == 4)
                {
                    for (int i = p.getLoc().getX() - 1; i > -1; i--)
                    {
                        
                        if (grid.get(p.getLoc().getY()).get(i).getPiece() == null ? true : grid.get(p.getLoc().getY()).get(i).getPiece().getPlayer() != turn)
                        {
                            ignore = new ArrayList<Piece>();
                            add = new ArrayList<Piece>();
                            ignore.add(p);
                            ignore.add(grid.get(p.getLoc().getY()).get(i).getPiece());
                            add.add(new Piece(p, new Point(i, p.getLoc().getY())));
                            if (!check(null, ignore, add))
                                return false;
                            if (grid.get(p.getLoc().getY()).get(i).getPiece() != null)
                                i = -1;
                        }
                        else
                            i = -1;
                    }
                    for (int i = p.getLoc().getX() + 1; i < 8; i++)
                    {
                        if (grid.get(p.getLoc().getY()).get(i).getPiece() == null ? true : grid.get(p.getLoc().getY()).get(i).getPiece().getPlayer() != turn)
                        {
                            ignore = new ArrayList<Piece>();
                            add = new ArrayList<Piece>();
                            ignore.add(p);
                            ignore.add(grid.get(p.getLoc().getY()).get(i).getPiece());
                            add.add(new Piece(p, new Point(i, p.getLoc().getY())));
                            if (!check(null, ignore, add))
                                return false;
                            if (grid.get(p.getLoc().getY()).get(i).getPiece() != null)
                                i = 8;
                        }
                        else
                            i = 8;
                    }
                    for (int i = p.getLoc().getY() - 1; i > -1; i--)
                    {
                        if (grid.get(i).get(p.getLoc().getX()).getPiece() == null ? true : grid.get(i).get(p.getLoc().getX()).getPiece().getPlayer() != turn)
                        {
                            ignore = new ArrayList<Piece>();
                            add = new ArrayList<Piece>();
                            ignore.add(p);
                            ignore.add(grid.get(i).get(p.getLoc().getX()).getPiece());
                            add.add(new Piece(p, new Point(p.getLoc().getX(), i)));
                            if (!check(null, ignore, add))
                                return false;
                            if (grid.get(i).get(p.getLoc().getX()).getPiece() != null)
                                i = -1;
                        }
                        else
                            i = -1;
                    }
                    for (int i = p.getLoc().getY() + 1; i < 8; i++)
                    {
                        if (grid.get(i).get(p.getLoc().getX()).getPiece() == null ? true : grid.get(i).get(p.getLoc().getX()).getPiece().getPlayer() != turn)
                        {
                            ignore = new ArrayList<Piece>();
                            add = new ArrayList<Piece>();
                            ignore.add(p);
                            ignore.add(grid.get(i).get(p.getLoc().getX()).getPiece());
                            add.add(new Piece(p, new Point(p.getLoc().getX(), i)));
                            if (!check(null, ignore, add))
                                return false;
                            if (grid.get(i).get(p.getLoc().getX()).getPiece() != null)
                                i = 8;
                        }
                        else
                            i = 8;
                    }
                }
                if (p.getType() == 2)
                {
                    for (int i = -2; i < 3; i += 4)
                        for (int k = -1; k < 2; k += 2)
                            if (p.getLoc().getX() + i > -1 && p.getLoc().getX() + i < 8 && p.getLoc().getY() + k > -1 && p.getLoc().getY() + k < 8 && (grid.get(p.getLoc().getY() + k).get(p.getLoc().getX() + i).getPiece() == null ? true : grid.get(p.getLoc().getY() + k).get(p.getLoc().getX() + i).getPiece().getPlayer() != turn))
                            {
                                ignore = new ArrayList<Piece>();
                                add = new ArrayList<Piece>();
                                ignore.add(p);
                                ignore.add(grid.get(p.getLoc().getY() + k).get(p.getLoc().getX() + i).getPiece());
                                add.add(new Piece(p, new Point(p.getLoc().getX() + i, p.getLoc().getY() + k)));
                                if (!check(null, ignore, add))
                                    return false;
                            }
                    for (int i = -1; i < 2; i += 2)
                        for (int k = -2; k < 3; k += 4)
                            if (p.getLoc().getX() + i > -1 && p.getLoc().getX() + i < 8 && p.getLoc().getY() + k > -1 && p.getLoc().getY() + k < 8 && (grid.get(p.getLoc().getY() + k).get(p.getLoc().getX() + i).getPiece() == null ? true : grid.get(p.getLoc().getY() + k).get(p.getLoc().getX() + i).getPiece().getPlayer() != turn))
                            {
                                ignore = new ArrayList<Piece>();
                                add = new ArrayList<Piece>();
                                ignore.add(p);
                                ignore.add(grid.get(p.getLoc().getY() + k).get(p.getLoc().getX() + i).getPiece());
                                add.add(new Piece(p, new Point(p.getLoc().getX() + i, p.getLoc().getY() + k)));
                                if (!check(null, ignore, add))
                                    return false;
                            }
                }
                if (p.getType() == 3 || p.getType() == 4)
                {
                    loopA: for (int i = 1; p.getLoc().getX() - i > -1 && p.getLoc().getY() - i > -1; i++)
                        if (grid.get(p.getLoc().getY() - i).get(p.getLoc().getX() - i).getPiece() == null ? true : grid.get(p.getLoc().getY() - i).get(p.getLoc().getX() - i).getPiece().getPlayer() != turn)
                        {
                            ignore = new ArrayList<Piece>();
                            add = new ArrayList<Piece>();
                            ignore.add(p);
                            ignore.add(grid.get(p.getLoc().getY() - i).get(p.getLoc().getX() - i).getPiece());
                            add.add(new Piece(p, new Point(p.getLoc().getX() - i, p.getLoc().getY() - i)));
                            if (!check(null, ignore, add))
                                return false;
                            if (grid.get(p.getLoc().getY() - i).get(p.getLoc().getX() - i).getPiece() == null)
                                break loopA;
                        }
                        else
                            break loopA;
                    loopA: for (int i = 1; p.getLoc().getX() + i < 8 && p.getLoc().getY() + i < 8; i++)
                        if (grid.get(p.getLoc().getY() + i).get(p.getLoc().getX() + i).getPiece() == null ? true : grid.get(p.getLoc().getY() + i).get(p.getLoc().getX() + i).getPiece().getPlayer() != turn)
                        {
                            ignore = new ArrayList<Piece>();
                            add = new ArrayList<Piece>();
                            ignore.add(p);
                            ignore.add(grid.get(p.getLoc().getY() + i).get(p.getLoc().getX() + i).getPiece());
                            add.add(new Piece(p, new Point(p.getLoc().getX() + i, p.getLoc().getY() + i)));
                            if (!check(null, ignore, add))
                                return false;
                            if (grid.get(p.getLoc().getY() + i).get(p.getLoc().getX() + i).getPiece() == null)
                                break loopA;
                        }
                        else
                            break loopA;
                    loopA: for (int i = 1; p.getLoc().getX() + i < 8 && p.getLoc().getY() - i > -1; i++)
                        if (grid.get(p.getLoc().getY() - i).get(p.getLoc().getX() + i).getPiece() == null ? true : grid.get(p.getLoc().getY() - i).get(p.getLoc().getX() + i).getPiece().getPlayer() != turn)
                        {
                            ignore = new ArrayList<Piece>();
                            add = new ArrayList<Piece>();
                            ignore.add(p);
                            ignore.add(grid.get(p.getLoc().getY() - i).get(p.getLoc().getX() + i).getPiece());
                            add.add(new Piece(p, new Point(p.getLoc().getX() + i, p.getLoc().getY() - i)));
                            if (!check(null, ignore, add))
                                return false;
                            if (grid.get(p.getLoc().getY() - i).get(p.getLoc().getX() + i).getPiece() == null)
                                break loopA;
                        }
                        else
                            break loopA;
                    loopA: for (int i = 1; p.getLoc().getX() - i > -1 && p.getLoc().getY() + i < 8; i++)
                        if (grid.get(p.getLoc().getY() + i).get(p.getLoc().getX() - i).getPiece() == null ? true : grid.get(p.getLoc().getY() + i).get(p.getLoc().getX() - i).getPiece().getPlayer() != turn)
                        {
                            ignore = new ArrayList<Piece>();
                            add = new ArrayList<Piece>();
                            ignore.add(p);
                            ignore.add(grid.get(p.getLoc().getY() + i).get(p.getLoc().getX() - i).getPiece());
                            add.add(new Piece(p, new Point(p.getLoc().getX() - i, p.getLoc().getY() + i)));
                            if (!check(null, ignore, add))
                                return false;
                            if (grid.get(p.getLoc().getY() + i).get(p.getLoc().getX() - i).getPiece() == null)
                                break loopA;
                        }
                        else
                            break loopA;
                }
                else if (p.getType() == 5)
                {
                    for (int i = p.getLoc().getY() - 1; i < p.getLoc().getY() + 2; i++)
                        if (i > -1 && i < 8)
                            for (int k = p.getLoc().getX() - 1; k < p.getLoc().getX() + 2; k++)
                                if (k > -1 && k < 8 && (i != p.getLoc().getY() || k != p.getLoc().getX()) && (grid.get(i).get(k).getPiece() == null ? true : grid.get(i).get(k).getPiece().getPlayer() != turn) && !check(new Point(k, i), new ArrayList<Piece>(), new ArrayList<Piece>()))
                                    return false;
                    if (turn == 1 && wCastleKing)
                    {
                        if (wCastleRookA)
                        {
                            boolean flag = false;
                            for (int i = 1; i < 4; i++)
                                if (grid.get(7).get(4 - i).getPiece() != null)
                                    flag = true;
                            if (!flag)
                            {
                                ignore = new ArrayList<Piece>();
                                add = new ArrayList<Piece>();
                                add.add(new Piece(grid.get(0).get(7).getPiece(), new Point(3, 7)));
                                ignore.add(grid.get(0).get(7).getPiece());
                                if (!check(new Point(2, 7), ignore, add))
                                    return false;
                            }
                        }
                        if (wCastleRookB)
                        {
                            boolean flag = false;
                            for (int i = 1; i < 3; i++)
                                if (grid.get(7).get(4 + i).getPiece() != null)
                                    flag = true;
                            if (!flag)
                            {
                                ignore = new ArrayList<Piece>();
                                add = new ArrayList<Piece>();
                                add.add(new Piece(grid.get(7).get(7).getPiece(), new Point(5, 7)));
                                ignore.add(grid.get(7).get(7).getPiece());
                                if (!check(new Point(6, 7), ignore, add))
                                    return false;
                            }
                        }
                    }
                    else if (turn == 2 && bCastleKing)
                    {
                        if (wCastleRookA)
                        {
                            boolean flag = false;
                            for (int i = 1; i < 4; i++)
                                if (grid.get(0).get(4 - i).getPiece() != null)
                                    flag = true;
                            if (!flag)
                            {
                                ignore = new ArrayList<Piece>();
                                add = new ArrayList<Piece>();
                                add.add(new Piece(grid.get(0).get(0).getPiece(), new Point(3, 0)));
                                ignore.add(grid.get(0).get(0).getPiece());
                                if (!check(new Point(2, 0), ignore, add))
                                    return false;
                            }
                        }
                        if (wCastleRookB)
                        {
                            boolean flag = false;
                            for (int i = 1; i < 3; i++)
                                if (grid.get(0).get(4 + i).getPiece() != null)
                                    flag = true;
                            if (!flag)
                            {
                                ignore = new ArrayList<Piece>();
                                add = new ArrayList<Piece>();
                                add.add(new Piece(grid.get(7).get(0).getPiece(), new Point(5, 0)));
                                ignore.add(grid.get(7).get(0).getPiece());
                                if (!check(new Point(6, 0), ignore, add))
                                    return false;
                            }
                        }
                    }
                }
            }
        }
        return true;
    }
    public ArrayList<ArrayList<Square>> convert() // to a diamond for displaying
    {
        ArrayList<ArrayList<Square>> newTiles = new ArrayList<ArrayList<Square>>();
        for (int y = 0; y < grid.size(); y++)
        {
            newTiles.add(new ArrayList<Square>()); // add row
            for (int h = y; h > -1; h--) // reading diagonal (southwest to northeast) for top half of map
                newTiles.get(newTiles.size() - 1).add(grid.get(h).get(y - h));
        }
        for (int x = 1; x < grid.size(); x++)
        {
            newTiles.add(new ArrayList<Square>()); // add row
            for (int v = x; v < grid.size(); v++)
                newTiles.get(newTiles.size() - 1).add(grid.get(grid.size() - (v - x + 1)).get(v));
        }
        return newTiles;
    }
}